# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import re
import unittest

import Namcap.version


class VersionTests(unittest.TestCase):
    def test_manpage(self):
        """Test that the manpage and program has the same version."""
        here = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(here, "..", "..", "namcap.1")) as f:
            first_line = f.readline()
        match = re.search(r'"namcap (.*?)"', first_line)
        assert match is not None
        self.assertEqual(match.group(1), Namcap.version.get_version())



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os

import Namcap.rules.rpath
from Namcap.tests.makepkg import MakepkgTest


class RpathTest(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_rpath
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
source=()
options=(!purge !zipman)
build() {
  cd "${srcdir}"
  echo "int main() { return 0; }" > main.c
  /usr/bin/gcc -o main.o -c main.c
  # core/binutils defaults to DT_RUNPATH since 2.39-4. Force old DT_RPATH here for testing.
  /usr/bin/ld -o main --disable-new-dtags --rpath /home/evil/lib main.o -lc
}
package() {
  install -D -m 755 "$srcdir/main" "$pkgdir/usr/bin/evilprogram"
}
"""

    def test_rpath_files(self):
        "Package with fancy RPATHs"
        pkgfile = "__namcap_test_rpath-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.rpath.package)
        self.assertEqual(r.errors, [("insecure-rpath %s %s", ("/home/evil/lib", "usr/bin/evilprogram"))])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import collections
import gzip
from typing import Any, TYPE_CHECKING
import os
import re
import subprocess
import sys

import pyalpm
import pycman.config


if TYPE_CHECKING:
    from .types import FormatArgs

_pyalpm_version_tuple = tuple(int(n) for n in pyalpm.version().split("."))
if _pyalpm_version_tuple < (0, 5):
    raise DeprecationWarning("pyalpm versions <0.5 are no longer supported")

pyalpm_handle = pycman.config.init_with_config("/etc/pacman.conf")

DEPENDS_RE = re.compile(r"([^<>=:]+)([<>]?=.*)?(: .*)?")
SODEPENDS_RE = re.compile(r"([^:]+)(: .*)?")


def strip_depend_info(value):
    """
    Strip all the depend version info off ('neon>=0.25.5-4' => 'neon'), keep soname versions.
    Also remove any trailing description (for optdepends)
    """
    if ".so=" in value:
        m = SODEPENDS_RE.match(value)
    else:
        m = DEPENDS_RE.match(value)
    if m is None:
        raise ValueError("Invalid dependency specification")
    return m.group(1)


class PacmanPackage(collections.abc.MutableMapping[str, Any]):
    strings = [
        "base",
        "name",
        "version",
        "desc",
        "url",
        "builddate",
        "packager",
        "install",
        "filename",
        "csize",
        "isize",
        "pkgfunction",
    ]
    equiv_vars = {
        "pkgname": "name",
        "md5sum": "md5sums",
        "sha1sum": "sha1sums",
        "depend": "depends",
        "pkgdesc": "desc",
        "size": "isize",
        "optdepend": "optdepends",
        "license": "licenses",
    }

    pkgbuild: list[str]

    @classmethod
    def canonical_varname(cls, varname):
        try:
            return cls.equiv_vars[varname]
        except KeyError:
            return varname

    def __init__(self, data=None, pkginfo=None, db=None):
        """
        A PacmanPackage object can be ininitialised from:
        * a dictionary (then its contents are updated accordingly)
        * the contents of a .PKGINFO file
        * the contents of a database entry or the output of parsepkgbuild
        """

        # Usual attributes
        self.is_split = False
        # a dictionary { package => [reasons why it is needed] }
        self.detected_deps: dict[str, list[tuple[str, FormatArgs]]] = collections.defaultdict(list)
        self._data = {}

        # Init from a dictionary
        if isinstance(data, dict):
            for k, v in data.items():
                self[k] = v

        # Parsing of .PKGINFO files from tarballs
        if isinstance(pkginfo, str):
            for i in pkginfo.splitlines():
                m = re.match(r"(.*) = (.*)", i)
                if m is not None:
                    lhs = m.group(1)
                    rhs = m.group(2)
                    if rhs != "":
                        self.setdefault(lhs, []).append(rhs)
        elif pkginfo is not None:
            raise TypeError("argument 'pkginfo' must be a string")

        # Parsing of database entries or parsepkgbuild output
        if isinstance(db, str):
            attrname = None
            if "\0" in db:
                self.is_split = True
                parts = db.split("\0")
                self.subpackages = [PacmanPackage(db=s) for s in parts[1:]]
                db = parts[0]
            for line in db.split("\n"):
                if line.startswith("%"):
                    attrname = line.strip("%").lower()
                elif line.strip() != "" and attrname:
                    self.setdefault(attrname, []).append(line)
        elif db is not None:
            raise TypeError("argument 'pkginfo' must be a string")

        # Cleanup
        self.process()

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, key):
        return self._data[self.canonical_varname(key)]

    def __setitem__(self, key, value):
        k = self.canonical_varname(key)
        self._data[k] = value

    def __contains__(self, key):
        return self.canonical_varname(key) in self._data

    def __delitem__(self, key):
        del self._data[self.canonical_varname(key)]

    def process_strings(self) -> None:
        """
        Turn all the instance properties listed in self.strings into strings instead of lists
        """
        for i in self.strings:
            if i in self:
                if isinstance(self[i], list):
                    self[i] = self[i][0]

    def clean_depends(self) -> None:
        """
        Strip all the depend version info off ('neon>=0.25.5-4' => 'neon').
        Also clean our optdepends and remove any trailing description.
        The original arrays are copied to orig_depends and orig_optdepends respectively.
        """

        if "depends" in self._data:
            self["orig_depends"] = self["depends"]
            self["depends"] = [strip_depend_info(d) for d in self["orig_depends"]]
        if "makedepends" in self._data:
            self["orig_makedepends"] = self["makedepends"]
            self["makedepends"] = [strip_depend_info(d) for d in self["orig_makedepends"]]
        if "optdepends" in self._data:
            self["orig_optdepends"] = self["optdepends"]
            self["optdepends"] = [strip_depend_info(d) for d in self["orig_optdepends"]]
        if "provides" in self._data:
            self["orig_provides"] = self["provides"]
            self["provides"] = [strip_depend_info(d) for d in self["orig_provides"]]

    def process(self) -> None:
        """
        After all the text processing happens, call this to sanitize the PacmanPackage object a bit
        """
        self.process_strings()
        self.clean_depends()

    def __repr__(self):
        if not self.is_split:
            return "PacmanPackage(%s)" % repr(self._data)
        else:
            children = ",".join(repr(p) for p in self.subpackages)
            return "PacmanPackage(%s,[%s])" % (repr(self._data), children)


def load_from_pkgbuild(path):
    # Load all the data like we normally would
    workingdir = os.path.dirname(path)
    if workingdir == "":
        workingdir = None
    filename = os.path.basename(path)
    process = subprocess.Popen(
        ["parsepkgbuild", filename], stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=workingdir
    )
    out_b, err_b = process.communicate()
    out = out_b.decode("utf-8", "ignore")
    err = err_b.decode("utf-8", "ignore")
    # this means parsepkgbuild returned an error, so we are not valid
    if process.returncode > 0:
        if out:
            print("Error:", out)
        if err:
            print("Error:", err, file=sys.stdout)
        return None
    ret = PacmanPackage(db=out)

    # Add a nice little .pkgbuild surprise
    pkgbuild = open(path, errors="ignore")
    ret.pkgbuild = pkgbuild.read().replace("\\\n", " ").splitlines()
    pkgbuild.close()

    return ret


def load_from_alpm(pmpkg):
    variables = [
        "name",
        "version",
        "conflicts",
        "url",
        "depends",
        "desc",
        "files",
        "groups",
        "has_scriptlet",
        "size",
        "licenses",
        "optdepends",
        "packager",
        "provides",
        "replaces",
    ]
    values = dict((v, getattr(pmpkg, v)) for v in variables)

    # arch is a list for PKGBUILDs, we do the same for tarball packages
    values["arch"] = [pmpkg.arch]
    # also drop md5sums for backed up files
    values["backup"] = [name for (name, md5) in pmpkg.backup]

    return PacmanPackage(data=values)


def load_from_tarball(path):
    try:
        p = pyalpm_handle.load_pkg(path)
    except pyalpm.error:
        return None

    return load_from_alpm(p)


def load_from_db(pkgname, dbname=None):
    if dbname is None:
        # default is loading local database
        db = pyalpm_handle.get_localdb()
    else:
        db = pyalpm_handle.register_syncdb(dbname, 0)
    p = db.get_pkg(pkgname)

    if p is None:
        p = lookup_provider(pkgname, db)
    if p is not None:
        p = load_from_alpm(p)
    return p


def load_testing_package(pkgname):
    "Loads the testing version of a package, None if not found."
    testing_dbs = [
        db for db in pyalpm_handle.get_syncdbs() if db.name in ("testing", "multilib-testing", "community-testing")
    ]
    for db in testing_dbs:
        p = db.get_pkg(pkgname)
        if p is not None:
            return load_from_alpm(p)


def get_installed_packages():
    return pyalpm_handle.get_localdb().pkgcache


def lookup_provider(pkgname, db):
    for pkg in db.pkgcache:
        stripped_provides = [strip_depend_info(d) for d in pkg.provides]
        if pkgname in stripped_provides:
            return pkg


def mtree_line(line):
    "returns head, {key:value}"
    # todo, un-hex the escaped chars
    head, _, kvs = line.partition(" ")
    kvs = dict(kv.split("=") for kv in kvs.split(" "))
    return head, kvs


def load_mtree(tar):
    "takes a tar object, returns (path, {attributes})"
    if ".MTREE" not in tar.getnames():
        raise StopIteration
    zfile = tar.extractfile(".MTREE")
    text = gzip.open(zfile).read().decode("utf-8")
    defaults = {}
    for line in text.split("\n"):
        if not line:
            continue
        if line.startswith("#"):
            continue
        head, kvs = mtree_line(line)
        if head == "/set":
            defaults = kvs
        attr = {}
        attr.update(defaults)
        attr.update(kvs)
        yield head, attr



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Check for py timestamps that are ahead of pyc/pyo timestamps
"""

import os

from Namcap.package import load_mtree
from Namcap.ruleclass import TarballRule


def _quick_filter(names):
    "can this package be skipped outright"
    if not names:
        return True
    found_py = any(n.endswith(".py") for n in names)
    found_pyc = any(n.endswith(".pyc") for n in names)
    found_pyo = any(n.endswith(".pyo") for n in names)
    if found_py and found_pyc:
        return False
    if found_py and found_pyo:
        return False
    return True


def _tar_timestamps(tar):
    "takes a tar object"
    return dict((m.name, m.mtime) for m in tar.getmembers())


def _mtree_timestamps(tar):
    "takes a tar object"
    return dict((h, a["time"]) for h, a in load_mtree(tar) if "time" in a)


def _generic_timestamps(tar):
    "works for mtree and tar"
    if ".MTREE" in tar.getnames():
        return _mtree_timestamps(tar)
    return _tar_timestamps(tar)


def _try_mtree(tar):
    "returns True if good, False if bad, None if N/A"
    if ".MTREE" not in tar.getnames():
        return None
    stamps = _mtree_timestamps(tar)
    if _quick_filter(stamps.keys()):
        return True
    return not _mtime_filter(stamps)


def _try_tar(tar):
    "returns True if good, False if bad"
    names = tar.getnames()
    if _quick_filter(names):
        return True
    mtimes = _tar_timestamps(tar)
    return not _mtime_filter(mtimes)


def _split_all(path):
    "like os.path.split but splits every directory"
    p2 = path
    dirs: list[str] = []
    while p2 and p2 != "/":
        p2, p3 = os.path.split(p2)
        dirs.insert(0, p3)
    # dirs.insert(0, '/')
    return dirs


def _source_py(path):
    "given a pyc/pyo, return the source path"
    if not path.endswith(".pyc") and not path.endswith(".pyo"):
        return None
    path = path[:-1]
    # handle py2
    if "__pycache__" not in path:
        return path
    # handle py3
    splitup = _split_all(path)
    if splitup[-2] != "__pycache__":
        return None
    splitup.pop(-2)
    f = splitup[-1]
    f = f.split(".")
    f.pop(-2)
    splitup[-1] = ".".join(f)
    return os.path.join(*splitup)


def _mtime_filter(mtimes):
    "return list of bad py file names"
    bad = []
    for name, mt2 in mtimes.items():
        if not name.endswith(".pyc") and not name.endswith(".pyo"):
            continue
        source_name = _source_py(name)
        if source_name not in mtimes:
            continue
        mt1 = mtimes[source_name]
        if mt1 > mt2:
            bad.append(source_name)
    return bad


class package(TarballRule):
    name = "py_mtime"
    description = "Check for py timestamps that are ahead of pyc/pyo timestamps"

    def analyze(self, pkginfo, tar):
        mtree_status = _try_mtree(tar)
        tar_status = _try_tar(tar)
        if mtree_status is False and tar_status:
            # mtree only
            self.warnings = [("py-mtime-mtree-warning", ())]
        elif not tar_status:
            # tar or both
            self.errors = [("py-mtime-tar-error", ())]
        self.infos = [("py-mtime-file-name %s", f[1:]) for f in _mtime_filter(_generic_timestamps(tar))]



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from Namcap.ruleclass import TarballRule


class package(TarballRule):
    name = "perllocal"
    description = "Verifies the absence of perllocal.pod."

    def analyze(self, pkginfo, tar):
        for i in tar.getnames():
            if i.endswith("perllocal.pod"):
                self.errors.append(("perllocal-pod-present %s", i))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from pathlib import Path
from tarfile import TarFile, TarInfo

from license_expression import BaseSymbol, LicenseSymbol, LicenseWithExceptionSymbol, get_spdx_licensing

from Namcap.package import PacmanPackage, load_from_db
from Namcap.ruleclass import TarballRule
from Namcap.util import is_debug


def get_license_canonicalized(license: str) -> str:
    """Get the canonicalized form of a license string

    This function may raise an Exception if it's not possible to derive any meaning from the input string
    """
    licensing = get_spdx_licensing()
    return str(licensing.parse(license, strict=True))


def strip_plus_from_license(license: BaseSymbol) -> BaseSymbol:
    """Remove a "+" suffix from the license symbol if it contains one

    The license_expression module does not support the SPDX "+" operator, but parses it as part of the license
    symbol instead. This function strips the operator from the symbol if it is found.

    License identifiers ending with "-only" or "-or-later" should not be used with the plus operator. If such
    license identifier is detected, the plus operator is left intact and this will be later reported as an
    unknown license symbol.
    """
    license_str = str(license)
    if license_str[-1] != "+" or license_str.startswith("LicenseRef-"):
        return license
    license_id = license_str[:-1]
    if license_id.endswith(("-only", "-or-later")):
        return license
    else:
        return LicenseSymbol(license_id)


def get_license_symbols(pkg_license: str) -> set[BaseSymbol]:
    """Extract all license symbols from a license string

    This function may raise an Exception if it's not possible to derive any meaning from the input string.
    This may be due to being unable to parse the string at all or if a license exception in the string is not a valid
    SPDX license exception identifier.
    """
    licensing = get_spdx_licensing()
    if (license_expression := licensing.parse(pkg_license, strict=True)) is None:
        raise ValueError("Empty license string")
    license_symbols: set[BaseSymbol] = set()
    for symbol in license_expression.symbols:
        if isinstance(symbol, LicenseWithExceptionSymbol):
            license, exception = list(symbol.decompose())
            license = strip_plus_from_license(license)
            license_symbols.add(LicenseWithExceptionSymbol(license, exception))
        else:
            license_symbols.add(strip_plus_from_license(symbol))
    return license_symbols


def get_common_spdx_license_identifiers() -> set[BaseSymbol]:
    """Get all common license identifiers (those that are provided system-wide and can be shared)"""
    common: list[BaseSymbol] = [
        LicenseSymbol(f"{x.stem}") for x in sorted(Path("/usr/share/licenses/spdx/").glob("*.txt")) if x.is_file()
    ]
    return set(common)


def get_common_spdx_license_exceptions() -> set[BaseSymbol]:
    """Get all common license exceptions (those that are provided system-wide and can be shared)"""
    common: list[BaseSymbol] = [
        LicenseSymbol(f"{x.stem}", is_exception=True)
        for x in sorted(Path("/usr/share/licenses/spdx/exceptions/").glob("*.txt"))
        if x.is_file()
    ]
    return set(common)


def get_known_spdx_license_identifiers() -> set[BaseSymbol]:
    """Get all known SPDX license identifiers"""
    all_spdx_licenses: list[BaseSymbol] = []

    with open("/usr/share/licenses/known_spdx_license_identifiers.txt") as file:
        while line := file.readline():
            all_spdx_licenses.append(LicenseSymbol(line.rstrip("\n")))

    return set(all_spdx_licenses)


def get_known_spdx_license_exceptions() -> set[BaseSymbol]:
    """Get all known SPDX license exceptions"""
    all_spdx_licenses: list[BaseSymbol] = []

    with open("/usr/share/licenses/known_spdx_license_exceptions.txt") as file:
        while line := file.readline():
            all_spdx_licenses.append(LicenseSymbol(line.rstrip("\n"), is_exception=True))

    return set(all_spdx_licenses)


def get_uncommon_license_symbols(
    licenses: set[BaseSymbol],
    known_licenses: set[BaseSymbol],
    known_exceptions: set[BaseSymbol],
    common_licenses: set[BaseSymbol],
    common_exceptions: set[BaseSymbol],
) -> set[BaseSymbol]:
    """Get the set of all uncommon licenses from a set of licenses

    This function compares against all uncommon, but known SPDX license symbols (those that require a custom file) or
    those prefixed with "LicenseRef-".
    When encountering LicenseWithExceptionSymbols, both the LicenseSymbol and the exception are checked.
    """
    uncommon_licenses = known_licenses - common_licenses
    uncommon_exceptions = known_exceptions - common_exceptions

    uncommon_symbols: list[BaseSymbol] = []
    for symbol in licenses:
        # if the symbol also carries an exception match license symbol and exception symbol separately
        if isinstance(symbol, LicenseWithExceptionSymbol):
            license, exception = list(symbol.decompose())
            if license in uncommon_licenses or str(license).startswith("LicenseRef-"):
                uncommon_symbols.append(license)
            if exception in uncommon_exceptions or str(exception).startswith("LicenseRef-"):
                uncommon_symbols.append(exception)
        else:
            if symbol in uncommon_licenses or str(list(symbol.decompose())[0]).startswith("LicenseRef-"):
                uncommon_symbols.append(symbol)

    return set(uncommon_symbols)


def get_unknown_license_symbols(
    licenses: set[BaseSymbol], known_licenses: set[BaseSymbol], known_exceptions: set[BaseSymbol]
) -> set[BaseSymbol]:
    """Get the set of all unknown license symbols from a set of licenses

    This function also evaluates LicenseWithExceptionSymbols and considers licenses and exceptions separately.
    """
    unknown_symbols: list[BaseSymbol] = []
    for symbol in licenses:
        # if the symbol also carries an exception match license symbol and exception symbol separately
        if isinstance(symbol, LicenseWithExceptionSymbol):
            license, exception = list(symbol.decompose())
            if license not in known_licenses:
                unknown_symbols.append(symbol)
            if exception not in known_exceptions:
                unknown_symbols.append(symbol)
        else:
            if symbol not in known_licenses:
                unknown_symbols.append(symbol)

    return set(unknown_symbols)


def get_symlink_target(member: TarInfo) -> str:
    """Return the symlink target of a TarInfo

    Absolute symlink targets will be returned as relative Path strings, as we are using them to search for files in
    TarFile objects (and all their members are relative).

    :raises: ValueError if the link target has an invalid amount of upward change dirs.
    """
    symlink_target: str
    if Path(member.linkname).is_absolute():
        symlink_target = str(Path(member.linkname.lstrip("/")))
    else:
        parents = Path(member.linkname).parts.count("..")
        reduce_name_by = -1 - parents

        # if we change dir upwards from the file name too often we may escape the original name and that's not valid
        if (len(Path(member.name).parts) + reduce_name_by) < 0:
            raise ValueError(f"Can not change dir upwards: {member.name} -> {member.linkname}")

        base_parts = Path(member.name).parts[0:reduce_name_by]
        add_parts = tuple(filter(lambda x: x != "..", Path(member.linkname).parts))
        symlink_target = str(Path("/".join(base_parts + add_parts)))

    return symlink_target


def package_license_files(tar: TarFile | None, pkgname: str) -> tuple[dict[str, bool], str | None]:
    """Return the license files referenced in a package and whether the license dir is a symlink"""
    license_dir_symlink = None
    files: dict[str, bool] = {}

    if not tar:
        return (files, license_dir_symlink)

    for member in tar.getmembers():
        # check if entire /usr/share/license/{pkgname}/ dir is a symlink
        if member.issym() and member.name == f"usr/share/licenses/{pkgname}":
            try:
                license_dir_symlink = get_symlink_target(member)
            except ValueError:
                continue
            else:
                break
        # the license is a file
        if (
            member.isfile()
            and not member.issym()
            and member.name.startswith(f"usr/share/licenses/{pkgname}/")
            and member.name != f"usr/share/licenses/{pkgname}"
        ):
            files[member.name] = True
        # the license is a symlink, we'll check later
        if member.issym() and member.name.startswith(f"usr/share/licenses/{pkgname}/"):
            try:
                files[get_symlink_target(member)] = False
            except ValueError:
                continue

    # check existence of all symlinked files
    for file, exists in [(file, exists) for (file, exists) in files.items() if not exists]:
        for member in tar.getmembers():
            if member.isfile() and not member.issym() and member.name == file:
                files[file] = True

    # if license dir is a symlink, add all files below the targeted license dir to files dict
    if license_dir_symlink:
        for member in tar.getmembers():
            if member.isfile() and not member.issym() and member.name.startswith(license_dir_symlink):
                files[member.name] = True

    return (files, license_dir_symlink)


class package(TarballRule):
    name = "licensepkg"
    description = "Verifies license is included in a package file"

    def analyze(self, pkginfo: PacmanPackage, tar: TarFile | None) -> None:
        # return early, as we do not check debug packages
        if is_debug(pkginfo):
            return

        # error and return early if there is no license
        if not pkginfo.get("license"):
            self.errors.append(("missing-license", ()))
            return

        # get canonicalized form of all license strings (and add errors for them)
        for license in pkginfo["license"]:
            try:
                canonicalized_license = get_license_canonicalized(license)
            except Exception:
                # we pass here, as we call get_license_symbols() later, which will raise as well
                pass
            else:
                if license != canonicalized_license:
                    self.errors.append(("license-statement-formatting %s %s", (license, canonicalized_license)))

        # get a combined set of all license symbols from all license strings
        # the license symbols may be of type LicenseSymbol (license) or LicenseWithExceptionSymbol (license + exception)
        license_symbols: set[BaseSymbol] = set()
        for license in pkginfo["license"]:
            try:
                new_license_symbols = get_license_symbols(license)
            except Exception:
                self.errors.append(("invalid-license-string %s", (license,)))
            else:
                license_symbols.update(new_license_symbols)

        known_licenses = get_known_spdx_license_identifiers()
        known_exceptions = get_known_spdx_license_exceptions()

        # check if any license (ignoring exception) symbols are unknown
        # (and add errors for them, if they are not prefixed with LicenseRef-)
        for license in get_unknown_license_symbols(license_symbols, known_licenses, known_exceptions):
            if not str(license).startswith("LicenseRef-"):
                self.errors.append(("unknown-spdx-license-identifier %s", (str(license),)))

        common_licenses = get_common_spdx_license_identifiers()
        common_exceptions = get_common_spdx_license_exceptions()

        # check whether there is a discrepancy between uncommon license symbols and license files found in the package
        uncommon_license_symbols = get_uncommon_license_symbols(
            license_symbols, known_licenses, known_exceptions, common_licenses, common_exceptions
        )
        if len(uncommon_license_symbols) == 0:
            return

        (pkg_licenses, license_dir_symlink) = package_license_files(tar, pkginfo["name"])
        licenses_in_pkg = len([(file, exists) for (file, exists) in pkg_licenses.items() if exists])
        licenses_outside_pkg = len([(file, exists) for (file, exists) in pkg_licenses.items() if not exists])

        # there are not enough license files in the package's license directory
        if licenses_in_pkg < len(uncommon_license_symbols) and licenses_outside_pkg == 0 and not license_dir_symlink:
            self.errors.append(
                (
                    "license-file-missing %s %s %s",
                    (
                        str(", ".join([str(id) for id in list(uncommon_license_symbols)])),
                        pkginfo["name"],
                        f"{licenses_in_pkg}/{len(uncommon_license_symbols)}",
                    ),
                )
            )

        # there are symlinks for license files that point to files in external packages
        if licenses_in_pkg == 0 and licenses_outside_pkg > 0 and not license_dir_symlink:
            outbound_licenses = [license for (license, exists) in pkg_licenses.items() if not exists]
            self.warnings.append(("license-file-in-external-pkg %s", (", ".join(outbound_licenses),)))

            for other_pkg in [load_from_db(name) for name in pkginfo["depends"]]:
                for outbound_license in outbound_licenses:
                    if outbound_license in other_pkg["files"]:
                        pkg_licenses[outbound_license] = True
                        # if there are no license files left to check, break early
                        if not [license for (license, exists) in pkg_licenses.items() if not exists]:
                            break

            outbound_licenses = [license for (license, exists) in pkg_licenses.items() if not exists]
            # there are symlinks to license files in other pkgs, but those files are missing
            # (or the package in question is not a dependency)
            for outbound_license in outbound_licenses:
                self.errors.append(
                    (
                        "license-file-missing-in-other-pkg %s",
                        (outbound_license,),
                    )
                )

        # the license dir is a symlink which points to another package
        if licenses_in_pkg == 0 and licenses_outside_pkg == 0 and license_dir_symlink:
            self.warnings.append(("license-dir-in-external-pkg %s", (license_dir_symlink,)))

            # try to figure out which other package contains the actual files
            other_pkg = (
                list(Path(license_dir_symlink.removeprefix("usr/share/licenses/")).parts)[0]
                if license_dir_symlink.startswith("usr/share/licenses/")
                else None
            )
            if other_pkg:
                if other_pkg not in pkginfo["depends"]:
                    self.errors.append(
                        (
                            "license-dir-target-pkg-not-in-depends %s %s %s",
                            (
                                other_pkg,
                                str(", ".join([str(id) for id in list(uncommon_license_symbols)])),
                                f"0/{len(uncommon_license_symbols)}",
                            ),
                        )
                    )
                    return

                other_pkg_info = load_from_db(other_pkg)
                for file, _, _ in other_pkg_info["files"]:
                    if file.startswith(license_dir_symlink) and file != license_dir_symlink:
                        licenses_outside_pkg += 1
            else:
                for other_pkg_info in [load_from_db(name) for name in pkginfo["depends"]]:
                    for file, _, _ in other_pkg_info["files"]:
                        if file.startswith(license_dir_symlink) and file != license_dir_symlink:
                            licenses_outside_pkg += 1

            # there are still some files missing
            if licenses_outside_pkg < len(uncommon_license_symbols):
                self.errors.append(
                    (
                        "license-dir-is-symlink-and-license-files-missing %s %s %s",
                        (
                            str(", ".join([str(id) for id in list(uncommon_license_symbols)])),
                            license_dir_symlink,
                            f"{licenses_outside_pkg}/{len(uncommon_license_symbols)}",
                        ),
                    )
                )



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import Namcap.rules.carch as module
from Namcap.tests.pkgbuild_test import PkgbuildTest


class NamcapSpecialArchTest(PkgbuildTest):
    pkgbuild1 = """
pkgname=mypackage
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  [[ $CARCH == x86_64 ]] && CFLAGS+="-m64"
  [ '$CARCH' = 'i686' ] && CFLAGS+="-m32"
  ./configure --prefix=/usr
  make
}
"""

    pkgbuild2 = """
pkgname=mypackage
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  [[ $CARCH == x86_64 ]] && CFLAGS+="-m64"
  [ '$CARCH' = 'i686' ] && CFLAGS+="-m32"
  ./configure --prefix=/usr
  make
}

package() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make DESTDIR="${pkgdir}" install
  cp foobar /usr/lib/i686/pkg/
}
"""

    test_valid = PkgbuildTest.valid_tests

    def preSetUp(self):
        self.rule = module.package

    def test_example1(self):
        r = self.run_on_pkg(self.pkgbuild1)
        self.assertEqual(r.errors, [])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_example2(self):
        r = self.run_on_pkg(self.pkgbuild2)
        self.assertEqual(r.errors, [])
        self.assertEqual(r.warnings, [("specific-host-type-used %s", ("i686",))])
        self.assertEqual(r.infos, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import re

from Namcap.ruleclass import TarballRule


class HookDependsRule(TarballRule):
    name = "hookdepends"
    description = "Check for redundant hook dependencies"
    subrules = [
        {
            "path": r"^usr/share/applications/.*\.desktop$",
            "dep": "desktop-file-utils",
        },
        {
            "path": r"^usr/share/mime$",
            "dep": "shared-mime-info",
        },
    ]

    def analyze(self, pkginfo, tar):
        names = [entry.name for entry in tar]
        for subrule in self.subrules:
            dep = subrule["dep"]
            if dep not in pkginfo["depends"]:
                continue
            pattern = re.compile(subrule["path"])
            if any(pattern.search(n) for n in names):
                self.warnings.append(("external-hooks-unneeded %s", (dep,)))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

"""Checks for invalid filenames."""

import string

from Namcap.ruleclass import TarballRule

VALID_CHARS = string.ascii_letters + string.digits + string.punctuation + " "


class package(TarballRule):
    name = "filenames"
    description = "Checks for invalid filenames."

    def analyze(self, pkginfo, tar):
        for i in tar.getnames():
            if not all(c in VALID_CHARS for c in i):
                self.warnings.append(("invalid-filename", i))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from Namcap.ruleclass import TarballRule


class package(TarballRule):
    name = "missingbackups"
    description = "Backup files listed in package should exist"

    def analyze(self, pkginfo, tar):
        if "backup" not in pkginfo or len(pkginfo["backup"]) == 0:
            return
        known_backups = set(pkginfo["backup"])
        found_files = set(tar.getnames())
        missing_backups = known_backups - found_files
        for backup in missing_backups:
            self.errors.append(("missing-backup-file %s", backup))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import re
from collections import defaultdict

import Namcap.package
from Namcap.ruleclass import TarballRule
from Namcap.util import is_elf

qml_path = "usr/lib/qt6/qml/"


def finddepends(modules):
    """
    Find packages owning a list of modules

    Returns:
      dependlist -- a dictionary { package => set(modules) }
      orphans -- the list of modules without owners
    """
    dependlist = defaultdict(set)
    foundlibs = set()

    for pkg in Namcap.package.get_installed_packages():
        for j, fsize, fmode in pkg.files:
            if j.startswith(qml_path) and j.endswith("/qmldir"):
                k = j.replace(qml_path, "").replace("/qmldir", "").replace("/", ".")
                if k in modules:
                    dependlist[pkg.name].add(k)
                    foundlibs.add(k)

    orphans = list(set(modules) - foundlibs)
    return dependlist, orphans


def get_imports(string, filename, modules):
    """
    Extract all QML imports from a file
    """
    import_pattern = r"^[ ]?import [\w.]+"
    imports = re.findall(import_pattern, string, re.MULTILINE)

    for m in imports:
        m = m.strip().replace("import ", "")
        modules[m].add(filename)
    return


class QmlDependencyRule(TarballRule):
    name = "qmldepends"
    description = "Checks QML dependencies"

    def analyze(self, pkginfo, tar):
        modules: dict[str, set[str]] = defaultdict(set)
        included_modules = []

        for entry in tar:
            if not entry.isfile():
                continue
            if entry.name.startswith(qml_path) and entry.name.endswith("/qmldir"):
                included_modules += [entry.name.replace(qml_path, "").replace("/qmldir", "").replace("/", ".")]
                continue
            if not entry.name.endswith(".qml") and not any(entry.name.startswith(d) for d in ["usr/bin", "usr/lib"]):
                continue
            f = tar.extractfile(entry)
            elf = is_elf(f)
            if not entry.name.endswith(".qml") and not elf:
                continue
            s = f.read().decode(errors="ignore")
            if elf and "libQt6Qml.so" not in s:
                # Does not embed QML, prevent false positives
                continue
            get_imports(s, entry.name, modules)
            f.close()

        for m in included_modules:
            modules.pop(m, None)

        dependlist, orphans = finddepends(modules)
        liblist = modules

        # Handle "no package associated" errors
        self.warnings.extend([("qml-module-no-package-associated %s %s", (i, str(list(liblist[i])))) for i in orphans])

        # Print QML module deps
        for pkg, libraries in dependlist.items():
            if isinstance(libraries, set):
                files = list(libraries)
                needing = set().union(*[liblist[lib] for lib in libraries])
                reasons = pkginfo.detected_deps.setdefault(pkg, [])
                reasons.append(("qml-modules-needed %s %s", (str(files), str(list(needing)))))
                self.infos.append(("qml-module-dependence %s in %s", (pkg, str(files))))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os

from elftools.elf.dynamic import DynamicSection
from elftools.elf.elffile import ELFFile

import Namcap.rules.sodepends
from Namcap.tests.makepkg import MakepkgTest


def get_soname(filename):
    with open(filename, "rb") as f:
        alpm = ELFFile(f)
        for section in alpm.iter_sections():
            if not isinstance(section, DynamicSection):
                continue
            for tag in section.iter_tags("DT_SONAME"):
                return tag.soname


class SoDependsTest(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_sodepends
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
source=()
options=(!purge !zipman)
build() {
  cd "${srcdir}"
  echo "int main() { return 0; }" > main.c
  gcc -o main main.c -Wl,--no-as-needed -lalpm
}
package() {
  install -D -m 755 "$srcdir/main" "$pkgdir/usr/bin/main"
}
"""

    def test_sodepends(self):
        "Package with missing pacman dependency"
        pkgfile = "__namcap_test_sodepends-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        alpm_filename = os.path.join("usr/lib", get_soname("/usr/lib/libalpm.so"))
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.sodepends.SharedLibsRule)
        self.assertEqual(
            pkg.detected_deps["pacman"], [("libraries-needed %s %s", (str([alpm_filename]), str(["usr/bin/main"])))]
        )
        e, w, i = Namcap.depends.analyze_depends(pkg)
        self.assertEqual(
            e,
            [
                (
                    "dependency-detected-not-included %s (%s)",
                    ("pacman", "libraries ['%s'] needed in files ['usr/bin/main']" % alpm_filename),
                )
            ],
        )
        self.assertEqual(w, [])



#!/usr/bin/env python3
# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import argparse
import os
import sys
import tarfile

import Namcap.depends
from Namcap.package import load_from_tarball, PacmanPackage
import Namcap.rules
import Namcap.tags
import Namcap.version


# Functions
def get_modules():
    """Return all possible modules (rules)"""
    return Namcap.rules.all_rules


def get_enabled_modules():
    """Return modules (rules) that should be used by default"""
    return dict(filter(lambda x: x[1].enable, get_modules().items()))


def open_package(filename):
    try:
        tar = tarfile.open(filename, "r")
        if ".PKGINFO" not in tar.getnames():
            tar.close()
            return None
    except OSError:
        if tar:
            tar.close()
        return None
    return tar


def show_messages(name, key, messages):
    colored_key = {
        "E": "\033[91mE\033[00m",
        "W": "\033[93mW\033[00m",
        "I": "\033[92mI\033[00m",
    }
    for msg in messages:
        if sys.stdout.isatty():
            print("%s %s: %s" % (name, colored_key[key], Namcap.tags.format_message(msg)))
        else:
            print("%s %s: %s" % (name, key, Namcap.tags.format_message(msg)))


def process_realpackage(package, modules):
    """Runs namcap checks over a package tarball"""
    pkgtar = open_package(package)

    if not pkgtar:
        print("Error: %s is empty or is not a valid package" % package)
        return 1

    pkginfo: PacmanPackage | None = load_from_tarball(package)
    if pkginfo is None:
        print(f"Error: Loading package from {package} failed")
        return 1

    # Loop through each one, load them apply if possible
    for i in modules:
        rule = get_modules()[i]()

        if isinstance(rule, Namcap.ruleclass.PkgInfoRule):
            rule.analyze(pkginfo, None)
        elif isinstance(rule, Namcap.ruleclass.PkgbuildRule):
            pass
        elif isinstance(rule, Namcap.ruleclass.TarballRule):
            rule.analyze(pkginfo, pkgtar)
        else:
            show_messages(pkginfo["name"], "E", [("error-running-rule %s", i)])

        # Output the three types of messages
        show_messages(pkginfo["name"], "E", rule.errors)
        show_messages(pkginfo["name"], "W", rule.warnings)
        if info_reporting:
            show_messages(pkginfo["name"], "I", rule.infos)

    # dependency analysis
    errs, warns, infos = Namcap.depends.analyze_depends(pkginfo)
    show_messages(pkginfo["name"], "E", errs)
    show_messages(pkginfo["name"], "W", warns)
    if info_reporting:
        show_messages(pkginfo["name"], "I", infos)


def process_pkginfo(pkginfo, modules):
    """Runs namcap checks of a single, non-split PacmanPackage object"""
    for i in modules:
        rule = get_modules()[i]()
        if isinstance(rule, Namcap.ruleclass.PkgInfoRule):
            rule.analyze(pkginfo, None)

        # Output the messages
        if "base" in pkginfo:
            name = "PKGBUILD (" + pkginfo["base"] + ")"
        else:
            name = "PKGBUILD (" + pkginfo["name"] + ")"
        show_messages(name, "E", rule.errors)
        show_messages(name, "W", rule.warnings)
        if info_reporting:
            show_messages(name, "I", rule.infos)


def process_pkgbuild(package, modules):
    """Runs namcap checks over a PKGBUILD"""
    # We might want to do some verifying in here... but really... isn't that
    # what pacman.load is for?
    pkginfo = Namcap.package.load_from_pkgbuild(package)

    if pkginfo is None:
        print("Error: %s is not a valid PKGBUILD" % package)
        return 1

    # apply global PKGBUILD rules
    for i in modules:
        rule = get_modules()[i]()
        if isinstance(rule, Namcap.ruleclass.PkgbuildRule):
            rule.analyze(pkginfo, package)
        # Output the messages
        if "base" in pkginfo:
            name = "PKGBUILD (" + pkginfo["base"] + ")"
        else:
            name = "PKGBUILD (" + pkginfo["name"] + ")"
        show_messages(name, "E", rule.errors)
        show_messages(name, "W", rule.warnings)
        if info_reporting:
            show_messages(name, "I", rule.infos)
    # apply per pkginfo rule
    for subpkg in pkginfo.subpackages if pkginfo.is_split else [pkginfo]:
        process_pkginfo(subpkg, modules)


# Main
modules = get_modules()
# Let's handle those options!
version = Namcap.version.get_version()

parser = argparse.ArgumentParser()
parser.add_argument("-L", "--list", action="store_true", help="List available rules")
parser.add_argument("-i", "--info", action="store_true", help="Prints information (debug) responses from rules")
parser.add_argument(
    "-m", "--machine-readable", action="store_true", help="Makes the output parseable (machine-readable)"
)
parser.add_argument("-t", "--tags", action="store", help="Use a custom tag file")
parser.add_argument("packages", nargs="*")
pargroup = parser.add_mutually_exclusive_group()
pargroup.add_argument(
    "-e",
    "--exclude",
    action="store",
    metavar="RULELIST",
    help="Don't apply RULELIST rules to the package (comma-separated)",
)
pargroup.add_argument(
    "-r",
    "--rules",
    action="store",
    metavar="RULELIST",
    help="Only apply RULELIST rules to the package (comma-separated)",
)
parser.add_argument("-v", "--version", action="version", version=version)
args = parser.parse_args()

if args.list:
    print("-" * 20 + " Namcap rule list " + "-" * 20)
    print(modules)
    for j in sorted(modules):
        print("%-20s: %s" % (j, modules[j].description))
    parser.exit(0)

if len(args.packages) == 0:
    print("Missing required argument packages", file=sys.stderr)
    parser.exit(2)

info_reporting = args.info
machine_readable = args.machine_readable
filename = args.tags
packages = args.packages

active_modules = {}

if args.rules:
    for rule in args.rules.split(","):
        if rule in modules:
            active_modules[rule] = modules[rule]
        else:
            print(f"Error: Rule '{rule}' does not exist")
            parser.exit(2)

if args.exclude:
    for rule in args.exclude.split(","):
        active_modules.update(modules)
        if rule in modules:
            active_modules.pop(rule)
        else:
            print(f"Error: Rule '{rule}' does not exist")
            parser.exit(2)

Namcap.tags.load_tags(filename=filename, machine=machine_readable)

# No rules selected?  Then use default selection
if len(active_modules) == 0:
    active_modules = get_enabled_modules()

# Go through each package, get the info, and apply the rules
for package in packages:
    if not os.access(package, os.R_OK):
        print("Error: Problem reading %s" % package)
        parser.print_usage()

    if os.path.isfile(package) and tarfile.is_tarfile(package):
        process_realpackage(package, active_modules)
    elif "PKGBUILD" in package:
        process_pkgbuild(package, active_modules)
    else:
        print("Error: %s not package or PKGBUILD" % package)



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from Namcap.ruleclass import TarballRule


class InfodirRule(TarballRule):
    name = "infodirectory"
    description = "Checks for info directory file."

    def analyze(self, pkginfo, tar):
        for i in tar.getnames():
            if i == "usr/share/info/dir":
                self.errors.append(("info-dir-file-present %s", i))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Check for ELF files to see if a package should be 'any' architecture
"""

import re

from Namcap.ruleclass import TarballRule
from Namcap.util import is_elf, is_static


class package(TarballRule):
    name = "anyelf"
    description = "Check for ELF files to see if a package should be 'any' architecture"

    def analyze(self, pkginfo, tar):
        supress_name = ["^mingw-"]
        if any(re.search(s, pkginfo["name"]) for s in supress_name):
            return
        found_elffiles = []

        for entry in tar.getmembers():
            if not entry.isfile():
                continue
            f = tar.extractfile(entry)
            # Ar files (static libs) are also architecture specific (FS#24854)
            if is_elf(f) or is_static(f):
                found_elffiles.append(entry.name)
            f.close()

        if pkginfo["arch"] and pkginfo["arch"][0] == "any":
            self.errors = [("elffile-in-any-package %s", i) for i in found_elffiles]
        else:
            if len(found_elffiles) == 0:
                self.warnings.append(("no-elffiles-not-any-package", ()))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

"This module contains rules that check coherence of split PKGBUILDs."

import Namcap.depends
from Namcap.ruleclass import PkgbuildRule


class PackageFunctionsRule(PkgbuildRule):
    """
    This rule checks for existence of package_* functions
    in split PKGBUILDs. It relies on the output of parsepkgbuild
    which gives the type of the package_xxx object in the PKGBUILD.
    """

    name = "splitpkgfunctions"
    description = "Checks that all package_* functions exist."

    def analyze(self, pkginfo, pkgbuild):
        if not pkginfo.is_split:
            return
        # All subpackages should have a valid package function
        for subpkg in pkginfo.subpackages:
            if subpkg["pkgfunction"] != "function":
                self.errors.append(("missing-pkgfunction %s", (subpkg["name"],)))


class SplitPkgMakedepsRule(PkgbuildRule):
    """
    This rule checks that the global makedepends of a split
    PKGBUILD cover all needed depends of individual children
    packages.
    """

    name = "splitpkgmakedeps"
    description = "Checks that a split PKGBUILD has enough makedeps."

    def analyze(self, pkginfo, pkgbuild):
        if not pkginfo.is_split:
            return
        # Find dependencies specified in makedepends
        global_deps = set()
        global_deps.update(pkginfo["names"])
        if "depends" in pkginfo:
            global_deps.update(pkginfo["depends"])
        if "makedepends" in pkginfo:
            global_deps.update(pkginfo["makedepends"])

        global_deps |= Namcap.depends.getcovered(global_deps)

        # Read dependencies specified in subpackages
        local_deps = set()
        for s in pkginfo.subpackages:
            if "depends" in s:
                local_deps.update(s["depends"])
            if "makedepends" in s:
                local_deps.update(s["makedepends"])

        if not local_deps.issubset(global_deps):
            missing = list(local_deps - global_deps)
            self.errors.append(("missing-makedeps %s", (str(sorted(missing)),)))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from Namcap.ruleclass import TarballRule
from Namcap.util import is_java


class JavaFiles(TarballRule):
    name = "javafiles"
    description = "Check for existence of Java classes or JARs"

    def analyze(self, pkginfo, tar):
        javas = []
        for entry in tar:
            # is it a regular file ?
            if not entry.isfile():
                continue
            # is it a JAR file ?
            if entry.name.endswith(".jar"):
                javas.append(entry.name)
                # self.infos.append( ('jar-file-found %s', entry.name) )
                continue
            # is it a CLASS file ?
            f = tar.extractfile(entry)
            if is_java(f):
                javas.append(entry.name)
                # self.infos.append( ('java-class-file-found %s', entry.name) )
            f.close()
        if javas:
            reasons = pkginfo.detected_deps.setdefault("java-runtime", [])
            reasons.append(("java-runtime-needed %s", ", ".join(javas)))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import unittest


def getTestSuite():
    loader = unittest.TestLoader()
    return loader.discover(start_dir=os.path.dirname(__file__))


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(getTestSuite())



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os

import Namcap.depends
import Namcap.rules.pathdepends
from Namcap.tests.makepkg import MakepkgTest


class HiColorIconsTest(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_hicoloricons
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=()
source=()
options=(!purge !zipman)
build() {
  true
}
package() {
  mkdir -p "${pkgdir}/usr/share/icons/hicolor/64x64"
}
"""

    def test_hicoloricons_file_exists(self):
        pkgfile = "__namcap_test_hicoloricons-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.pathdepends.PathDependsRule)
        e, w, i = Namcap.depends.analyze_depends(pkg)
        self.assertEqual(pkg.detected_deps, {"hicolor-icon-theme": [("hicolor-icon-theme-needed-for-hicolor-dir", ())]})
        self.assertEqual(
            set(r.errors + e),
            set(
                [
                    (
                        "dependency-detected-not-included %s (%s)",
                        ("hicolor-icon-theme", "needed for hicolor theme hierarchy"),
                    ),
                ]
            ),
        )
        self.assertEqual(r.warnings + w, [])
        self.assertEqual(r.infos + i, [("depends-by-namcap-sight depends=(%s)", ("hicolor-icon-theme",))])

    valid_pkgbuild = """
pkgname=__namcap_test_hicoloricons
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('hicolor-icon-theme')
source=()
options=(!purge !zipman)
build() {
  true
}
package() {
  mkdir -p "${pkgdir}/usr/share/icons/hicolor/64x64"
}
"""

    def test_hicoloricons_normal(self):
        pkgfile = "__namcap_test_hicoloricons-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.valid_pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.pathdepends.PathDependsRule)
        e, w, i = Namcap.depends.analyze_depends(pkg)
        self.assertEqual(pkg.detected_deps, {"hicolor-icon-theme": [("hicolor-icon-theme-needed-for-hicolor-dir", ())]})
        self.assertEqual(r.errors + e, [])
        self.assertEqual(r.warnings + w, [])
        self.assertEqual(
            r.infos + i,
            [
                ("dependency-detected-satisfied %s (%s)", ("hicolor-icon-theme", "needed for hicolor theme hierarchy")),
                ("depends-by-namcap-sight depends=(%s)", ("hicolor-icon-theme",)),
            ],
        )



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

"""Checks dependencies on programs specified in shebangs."""

import shutil

import Namcap.package
from Namcap.ruleclass import TarballRule
from Namcap.util import is_script, script_type


def scanshebangs(fileobj, filename, scripts):
    """
    Scan a file for shebang and stores the interpreter name.

    Stores
      scripts -- a dictionary { program => set(scripts) }
    """

    # test magic bytes
    if not is_script(fileobj):
        return
    # process shebang line
    cmd = script_type(fileobj)
    if cmd is not None:
        assert isinstance(cmd, str)
        scripts.setdefault(cmd, set()).add(filename)


def findowners(scriptlist):
    """
    Find owners for executables.

    Returns:
      pkglist -- a dictionary { package => set(programs) }
      orphans -- a set of scripts not found
    """

    pkglist: dict[str, set[str]] = {}
    scriptfound = set()

    for s in scriptlist:
        out = shutil.which(s)
        if not out:
            continue

        # strip leading slash
        scriptpath = out.lstrip("/")
        for pkg in Namcap.package.get_installed_packages():
            pkg_files = [fname for fname, fsize, fmode in pkg.files]
            if scriptpath in pkg_files:
                pkglist.setdefault(pkg.name, set()).add(s)
                scriptfound.add(s)

    orphans = list(set(scriptlist) - scriptfound)
    return pkglist, orphans


class ShebangDependsRule(TarballRule):
    name = "shebangdepends"
    description = "Checks dependencies semi-smartly."

    def analyze(self, pkginfo, tar):
        scriptlist: dict[str, set[str]] = {}

        for entry in tar:
            if not entry.isfile():
                continue
            f = tar.extractfile(entry)
            scanshebangs(f, entry.name, scriptlist)
            f.close()

        # find packages owning interpreters
        pkglist, orphans = findowners(scriptlist)
        for dep, progs in pkglist.items():
            if not isinstance(progs, set):
                continue
            needing = set().union(*[scriptlist[prog] for prog in progs])
            reasons = pkginfo.detected_deps.setdefault(dep, [])
            reasons.append(("programs-needed %s %s", (str(list(progs)), str(list(needing)))))

        # Do the script handling stuff
        for i, v in scriptlist.items():
            files = list(v)
            self.infos.append(("script-link-detected %s in %s", (i, str(files))))

        # Handle "no package associated" errors
        self.warnings.extend([("library-no-package-associated %s %s", (i, str(list(scriptlist[i])))) for i in orphans])

        # Check for packages in testing
        for i in scriptlist:
            p = Namcap.package.load_testing_package(i)
            q = Namcap.package.load_from_db(i)
            if p is not None and q is not None and p["version"] == q["version"]:
                self.warnings.append(("dependency-is-testing-release %s", (i,)))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import re

from Namcap.ruleclass import TarballRule


class FHSRule(TarballRule):
    name = "directoryname"
    description = "Checks for standard directories."

    def analyze(self, pkginfo, tar):
        valid_paths = [
            "etc/",
            "opt/",
            "lib/modules",
            "usr/bin/",
            "usr/include/",
            "usr/lib/",
            "usr/lib32/",
            "usr/sbin/",
            "usr/share/",
            "usr/src/",
            "var/cache/",
            "var/lib/",
            "var/log/",
            "var/opt/",
            "var/spool/",
            "var/state/",
            ".PKGINFO",
            ".INSTALL",
            ".CHANGELOG",
            ".MTREE",
            ".BUILDINFO",
        ]
        forbidden_paths = ["tmp/", "var/tmp/", "run/", "var/run/", "var/lock/"]
        custom_valid = {
            r"^mingw-": [
                "usr/x86_64-w64-mingw32/lib/",
                "usr/x86_64-w64-mingw32/bin/",
                "usr/x86_64-w64-mingw32/include/",
                "usr/i686-w64-mingw32/lib/",
                "usr/i686-w64-mingw32/bin/",
                "usr/i686-w64-mingw32/include/",
            ],
        }
        for pattern in custom_valid:
            if re.search(pattern, pkginfo["name"]):
                valid_paths.extend(custom_valid[pattern])
        for entry in tar.getmembers():
            name = os.path.normpath(entry.name)
            if entry.isdir():
                name += "/"

            # check for files in wrong dirs, directory itself will be
            # caught by emptydirs rule
            if name in forbidden_paths:
                continue
            bad_dirs = (name.startswith(dirname) for dirname in forbidden_paths)
            if any(bad_dirs):
                self.errors.append(("file-in-temporary-dir %s", name))
                continue

            # matches directory names or parent directories
            good_dirs = (name.startswith(dirname) or dirname.startswith(name) for dirname in valid_paths)
            if not any(good_dirs):
                self.warnings.append(("file-in-non-standard-dir %s", name))


class FHSManpagesRule(TarballRule):
    name = "fhs-manpages"
    description = "Verifies correct installation of man pages"

    def analyze(self, pkginfo, tar):
        gooddir = "usr/share/man"
        bad_dir = "usr/man"
        for i in tar.getmembers():
            if not i.isfile():
                continue
            if i.name.startswith(gooddir):
                continue
            if i.name.startswith(bad_dir):
                self.errors.append(("non-fhs-man-page %s", i.name))
                continue
            # Check everything else to see if it has a 'man' path component
            if "man" in i.name.split(os.sep):
                self.warnings.append(("potential-non-fhs-man-page %s", i.name))


class FHSInfoPagesRule(TarballRule):
    name = "fhs-infopages"
    description = "Verifies correct installation of info pages"

    def analyze(self, pkginfo, tar):
        for i in tar.getmembers():
            if not i.isfile():
                continue
            if i.name.startswith("usr/share/info"):
                continue
            if i.name.startswith("usr/info"):
                self.errors.append(("non-fhs-info-page %s", i.name))
                continue
            if "info" in i.name.split(os.sep):
                self.warnings.append(("potential-non-fhs-info-page %s", i.name))


class RubyPathsRule(TarballRule):
    name = "rubypaths"
    description = "Verifies correct usage of folders by ruby packages"

    def analyze(self, pkginfo, tar):
        for i in tar.getmembers():
            if i.name.startswith("usr/lib/ruby/site_ruby"):
                self.warnings.append(("site-ruby", ()))
                return



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from typing import Any, TypeAlias

FormatArgs: TypeAlias = tuple[Any, ...]

Diagnostic: TypeAlias = tuple[str, FormatArgs]



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from Namcap.ruleclass import PkgInfoRule


class package(PkgInfoRule):
    name = "pkgnameindesc"
    description = "Verifies if the package name is included on package description"

    def analyze(self, pkginfo, tar):
        if "name" in pkginfo and "desc" in pkginfo:
            if pkginfo["name"].lower() in pkginfo["desc"].lower().split():
                self.warnings.append(("pkgname-in-description", ()))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from Namcap.tests.pkgbuild_test import PkgbuildTest
from Namcap.rules.missingvars import ChecksumsRule, DescriptionRule, TagsRule


class NamcapChecksumTest(PkgbuildTest):
    pkgbuild1 = """
# Maintainer: Arch Linux <archlinux@example.com>
# Contributor: Arch Linux <archlinux@example.com>

pkgname=mypackage
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
options=('!libtool')
source=(ftp://ftp.example.com/pub/mypackage-0.1.tar.gz)
md5sums=('look-this-is-an-invalid-checksum')

build() {
  true
}

package() {
  true
}
"""
    pkgbuild2 = """
# Maintainer: Arch Linux <archlinux@example.com>
# Contributor: Arch Linux <archlinux@example.com>

pkgname=mypackage
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
options=('!libtool')
source=(ftp://ftp.example.com/pub/mypackage-0.1.tar.gz
		nice-patch.diff)
md5sums=('abcdefabcdef12345678901234567890')

build() {
  true
}

package() {
  true
}
"""

    # packages with no sources (FS#23258)
    pkgbuild_no_sources = """
# Maintainer: Arch Linux <archlinux@example.com>
# Contributor: Arch Linux <archlinux@example.com>

pkgname=mypackage
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc' 'pacman')
options=('!libtool')
md5sums=('1234567890abcdef1234567890abcdef')

build() {
wget http://www.example.com/sources.tar.gz
cd "${srcdir}"/${pkgname}-${pkgver}
./configure --prefix=/usr
make
}

package() {
cd "${srcdir}"/${pkgname}-${pkgver}
make DESTDIR="${pkgdir}" install
}
"""

    # package using CARCH (FS#32568)
    pkgbuild_sha512 = """
pkgname=chromium-libpdf
pkgver=24.0.1312.5
pkgrel=1
pkgdesc="PDF Viewer for Chromium"
url="http://wiki.archlinux.org/index.php/Chromium#libpdf.so"
arch=("i686" "x86_64")
license=('custom:chrome')
makedepends=('lynx')
depends=('chromium')
if [ "$CARCH" = "i686" ]; then
	_arch="i386"
	sha512sums=('31f63cd859203071e3b2a9b608bc80c67cac22f56c7de31d809f8c07cba6e076701c9649107db28a2fbef5ce1bf843e93b343a132058c3784d13a98492aa7bec')
elif [ "$CARCH" = "x86_64" ]; then
	_arch="amd64"
	sha512sums=('1989976e6ca93099c92957e44948d7b73a6f9594304b9fe5313ff44c28c5261852919f2b5dacb49724c77d98667cc3991f8aaef9d3d9b2277eb375293c4a0395')
fi
source=(https://dl.google.com/linux/direct/google-chrome-unstable_current_$_arch.deb)

build() {
}
"""

    # package using "SKIP", new in pacman 4.1 (FS#34647)
    pkgbuild_skip = """
pkgname=youtube-dl
pkgver=2013.02.25
pkgrel=1
pkgdesc="A small command-line program to download videos from YouTube.com and a few more sites"
arch=('any')
url="http://rg3.github.com/youtube-dl/"
license=('custom')
depends=('python')
makedepends=('python-distribute')
source=(http://youtube-dl.org/downloads/${pkgver}/${pkgname}-${pkgver}.tar.gz{,.sig})
sha1sums=('SKIP' 'SKIP')

package() {
}
"""

    test_valid = PkgbuildTest.valid_tests

    def preSetUp(self):
        self.rule = ChecksumsRule

    def test_example1(self):
        # Example 1
        r = self.run_on_pkg(self.pkgbuild1)
        self.assertEqual(r.errors, [("improper-checksum %s %s", ("md5sums", "look-this-is-an-invalid-checksum"))])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_example2(self):
        # Example 2
        r = self.run_on_pkg(self.pkgbuild2)
        self.assertEqual(r.errors, [("not-enough-checksums %s %i needed", ("md5sums", 2))])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_example_no_sources(self):
        # Example with no sources (FS #23259)
        r = self.run_on_pkg(self.pkgbuild_no_sources)
        self.assertEqual(r.errors, [("too-many-checksums %s %i needed", ("md5sums", 0))])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_example_valid(self):
        for p in [self.pkgbuild_sha512, self.pkgbuild_skip]:
            r = self.run_on_pkg(p)
            self.assertEqual(r.errors, [])
            self.assertEqual(r.warnings, [])
            self.assertEqual(r.infos, [])


class NamcapMaintainerTagTest(PkgbuildTest):
    pkgbuild1 = """
pkgname=mypackage
pkgver=1.0
pkgrel=1
pkgdesc="The foobar program"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
options=('!libtool')
source=(ftp://ftp.example.com/pub/mypackage-0.1.tar.gz)
md5sums=('abcdefabcdef12345678901234567890')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make
}

package() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make DESTDIR="${pkgdir}" install
}
"""

    test_valid = PkgbuildTest.valid_tests

    def preSetUp(self):
        self.rule = TagsRule

    def test_example1(self):
        # Example 1
        r = self.run_on_pkg(self.pkgbuild1)
        self.assertEqual(r.errors, [])
        self.assertEqual(r.warnings, [("missing-maintainer", ())])
        self.assertEqual(r.infos, [("missing-contributor", ())])


class NamcapPkgdesc(PkgbuildTest):
    pkgbuild1 = """
# Maintainer: Arch Linux <archlinux@example.com>
# Contributor: Arch Linux <archlinux@example.com>

pkgbase=base
pkgname=(foo bar)
pkgver=1.0
pkgrel=1
arch=('i686' 'x86_64')
depends=('glibc')
license=('GPL-3.0-or-later')
url="https://archlinux.org"
source=(ftp://ftp.example.com/pub/mypackage-0.1.tar.gz)
md5sums=('abcdefabcdef12345678901234567890')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make
}

package_foo() {
  depends=(glibc)

  cd "${srcdir}"/${pkgname}-${pkgver}
  make DESTDIR="${pkgdir}" install
}

package_bar() {
  depends=(glibc)

  cd "${srcdir}"/${pkgname}-${pkgver}
  make DESTDIR="${pkgdir}" install
}
"""

    pkgbuild2 = """
# Maintainer: Arch Linux <archlinux@example.com>
# Contributor: Arch Linux <archlinux@example.com>

pkgname=foo
pkgver=1.0
pkgrel=1
arch=('i686' 'x86_64')
depends=('glibc')
license=('GPL-3.0-or-later')
url="https://archlinux.org"
source=(ftp://ftp.example.com/pub/mypackage-0.1.tar.gz)
md5sums=('abcdefabcdef12345678901234567890')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make
}

package() {

  cd "${srcdir}"/${pkgname}-${pkgver}
  make DESTDIR="${pkgdir}" install
}
"""

    pkgbuild3 = """
# Maintainer: Arch Linux <archlinux@example.com>
# Contributor: Arch Linux <archlinux@example.com>

pkgname=foo
pkgver=1.0
pkgrel=1
arch=('i686' 'x86_64')
depends=('glibc')
pkgdesc='this is the package foo'
license=('GPL-3.0-or-later')
url="https://archlinux.org"
source=(ftp://ftp.example.com/pub/mypackage-0.1.tar.gz)
md5sums=('abcdefabcdef12345678901234567890')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make
}

package() {

  cd "${srcdir}"/${pkgname}-${pkgver}
  make DESTDIR="${pkgdir}" install
}
"""

    test_valid = PkgbuildTest.valid_tests

    def preSetUp(self):
        self.rule = DescriptionRule

    def test_example1(self):
        # Example 1
        r = self.run_on_pkg(self.pkgbuild1)
        self.assertEqual(r.errors, [("missing-description %s", ("foo",)), ("missing-description %s", ("bar",))])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_example2(self):
        r = self.run_on_pkg(self.pkgbuild2)
        self.assertEqual(r.errors, [("missing-description", ())])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_example3(self):
        r = self.run_on_pkg(self.pkgbuild3)
        self.assertEqual(r.errors, [])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os

import Namcap.rules.symlink
from Namcap.tests.makepkg import MakepkgTest


class SymlinkTest(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_symlink
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
source=()
options=(!purge !zipman)
backup=(etc/imaginary_file.conf)
build() {
  true
}
package() {
  mkdir -p "${pkgdir}/usr/share"
  touch "${pkgdir}/usr/share/somedata"
  ln -s nofile "${pkgdir}/usr/share/somelink"
  ln -s ../nofile "${pkgdir}/usr/share/somelink2"
  ln -s //usr/share/somedata "${pkgdir}/usr/share/validlink"
  ln -s ../share/somedata "${pkgdir}/usr/share/validlink2"
  ln -s /usr/include/math.h "${pkgdir}/usr/share/deplink"
}
"""

    def test_symlink_files(self):
        "Package with a dangling symlink"
        pkgfile = "__namcap_test_symlink-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.symlink.package)
        self.assertEqual(
            set(r.errors),
            set(
                [
                    ("dangling-symlink %s points to %s", ("usr/share/somelink", "nofile")),
                    ("dangling-symlink %s points to %s", ("usr/share/somelink2", "../nofile")),
                ]
            ),
        )
        self.assertEqual(r.warnings, [])
        self.assertEqual(
            set(r.infos),
            set(
                [
                    ("symlink-found %s points to %s", ("usr/share/somelink", "nofile")),
                    ("symlink-found %s points to %s", ("usr/share/somelink2", "../nofile")),
                    ("symlink-found %s points to %s", ("usr/share/validlink", "//usr/share/somedata")),
                    ("symlink-found %s points to %s", ("usr/share/validlink2", "../share/somedata")),
                    ("symlink-found %s points to %s", ("usr/share/deplink", "/usr/include/math.h")),
                ]
            ),
        )



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import shutil
import subprocess
import sys
import tarfile
import tempfile
import unittest

import Namcap.package

makepkg_conf = """
DLAGENTS=('ftp::/usr/bin/wget -c --passive-ftp -t 3 --waitretry=3 -O %%o %%u'
          'http::/usr/bin/wget -c -t 3 --waitretry=3 -O %%o %%u'
          'https::/usr/bin/wget -c -t 3 --waitretry=3 --no-check-certificate -O %%o %%u'
          'rsync::/usr/bin/rsync -z %%u %%o'
          'scp::/usr/bin/scp -C %%u %%o')
CARCH="%(arch)s"
CHOST="%(arch)s-pc-linux-gnu"
CFLAGS="-march=%(arch)s -mtune=generic -O2 -pipe"
CXXFLAGS="-march=%(arch)s -mtune=generic -O2 -pipe"
LDFLAGS="-Wl,--hash-style=gnu -Wl,--as-needed"
BUILDENV=(fakeroot !distcc color !ccache)
OPTIONS=(strip docs libtool emptydirs zipman purge)
INTEGRITY_CHECK=(md5)
STRIP_BINARIES="--strip-all"
STRIP_SHARED="--strip-unneeded"
STRIP_STATIC="--strip-debug"
MAN_DIRS=({usr{,/local}{,/share},opt/*}/{man,info})
DOC_DIRS=(usr/{,local/}{,share/}{doc,gtk-doc} opt/*/{doc,gtk-doc})
STRIP_DIRS=(bin lib sbin usr/{bin,lib,sbin,local/{bin,lib,sbin}} opt/*/{bin,lib,sbin})
PURGE_TARGETS=(usr/{,share}/info/dir .packlist *.pod)
PKGEXT='.pkg.tar.xz'
SRCEXT='.src.tar.gz'
"""


class MakepkgTest(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.arch = os.uname()[-1]
        with open(os.path.join(self.tmpdir, "makepkg.conf"), "w") as f:
            f.write(makepkg_conf % {"arch": self.arch})

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def run_makepkg(self):
        pwd = os.getcwd()
        os.chdir(self.tmpdir)
        environment = os.environ
        environment.update({"MAKEPKG_CONF": os.path.join(self.tmpdir, "makepkg.conf")})
        p = subprocess.Popen(["makepkg", "-f", "-d"], env=environment, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out1, out2 = p.communicate()
        if p.returncode != 0:
            sys.stderr.buffer.write(out1)
            sys.stderr.buffer.write(out2)
            raise RuntimeError("makepkg returned an error")
        os.chdir(pwd)

    def run_rule_on_tarball(self, filename, rule):
        # process PKGINFO
        pkg = Namcap.package.load_from_tarball(filename + ".xz")

        tar = tarfile.open(filename + ".xz")
        r = rule()
        r.analyze(pkg, tar)
        tar.close()
        return pkg, r



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os

import Namcap.rules.lotsofdocs
from Namcap.tests.makepkg import MakepkgTest


class LotsOfDocsTest(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_lotsofdocs
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
source=()
options=(!purge !zipman)
build() {
  true
}
package() {
  mkdir -p "${pkgdir}/usr/share/doc"
  dd if=/dev/zero of="${pkgdir}/usr/share/doc/somefile" bs=1k count=500
}
"""

    def test_lotsofdocs(self):
        "Empty package with a 500kB file in /usr/share/doc"
        pkgfile = "__namcap_test_lotsofdocs-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.lotsofdocs.package)
        self.assertEqual(r.errors, [])
        tag, value = r.warnings[0]
        self.assertEqual(tag, "lots-of-docs %f")
        self.assertGreater(value, (50,))
        self.assertEqual(r.infos, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from itertools import product

from Namcap.ruleclass import PkgbuildRule


class package(PkgbuildRule):
    name = "extravars"
    description = "Verifies that extra variables start with an underscore"

    def analyze(self, pkginfo, tar):
        carch_vars = [
            "checkdepends",
            "conflicts",
            "depends",
            "makedepends",
            "optdepends",
            "provides",
            "replaces",
            "source",
            "md5sums",
            "sha224sums",
            "sha1sums",
            "sha256sums",
            "sha384sums",
            "sha512sums",
            "b2sums",
        ]
        stdvars = [
            "arch",
            "license",
            "backup",
            "noextract",
            "pkgname",
            "pkgbase",
            "pkgver",
            "pkgrel",
            "epoch",
            "pkgdesc",
            "groups",
            "url",
            "install",
            "changelog",
            "options",
            "validpgpkeys",
        ] + carch_vars
        if "arch" in pkginfo and pkginfo["arch"] != ["any"]:
            stdvars.extend(v + "_" + a for v, a in product(carch_vars, pkginfo["arch"]))
        for varname in pkginfo["setvars"]:
            if varname.islower() and varname not in stdvars and not varname.startswith("_"):
                self.warnings.append(("extra-var-begins-without-underscore %s", varname))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

"""Checks dependencies semi-smartly."""

from typing import TYPE_CHECKING

import Namcap.tags
from Namcap import package

if TYPE_CHECKING:
    from .types import Diagnostic


def single_covered(depend):
    "Returns full coverage tree of one package, with loops broken"
    covered = set()
    todo = set([depend])
    while todo:
        i = todo.pop()
        covered.add(i)
        pac = package.load_from_db(i)
        if pac is None:
            continue
        todo |= set(pac["depends"]) - covered

    return covered - set([depend])


def getcovered(dependlist):
    """
    Returns full coverage tree set, without packages
    from self-loops (iterable of package names)
    """

    covered = set()
    for d in dependlist:
        covered |= single_covered(d)
    return covered


def getprovides(depends):
    provides: dict[str, set[str]] = {}
    for i in depends:
        provides[i] = set()
        pac = package.load_from_db(i)
        if pac is None:
            continue
        if not pac["provides"]:
            continue
        provides[i].update(pac["provides"])
    return provides


def analyze_depends(pkginfo):
    errors: list[Diagnostic] = []
    warnings: list[Diagnostic] = []
    infos: list[Diagnostic] = []

    # compute needed dependencies
    dependlist = set(pkginfo.detected_deps.keys())

    # Find all the covered dependencies from the PKGBUILD
    pkginfo.setdefault("depends", [])
    explicitdepend = set(pkginfo["depends"])
    implicitdepend = getcovered(explicitdepend)

    # Include the optdepends from the PKGBUILD
    pkginfo.setdefault("optdepends", [])
    optdepend = set(pkginfo["optdepends"])
    implicit_optdepend = getcovered(optdepend)

    # Get the provides so we can reference them later
    # smartprovides : depend => (packages provided by depend)
    smartprovides = getprovides(dependlist)

    # The set of all provides for detected dependencies
    allprovides = set()
    for plist in smartprovides.values():
        allprovides |= plist

    # Common deps
    for duplicated_optdepend in explicitdepend & optdepend:
        errors.append(("dependency-duplicated-optdepend %s", (duplicated_optdepend,)))
    for satisfied_optdepend in implicitdepend & optdepend:
        infos.append(("dependency-satisfied-optdepend %s", (satisfied_optdepend,)))

    # Do the actual message outputting stuff
    for i in dependlist:
        # compute dependency reason
        reasons = pkginfo.detected_deps[i]
        reason_strings = [Namcap.tags.format_message(reason) for reason in reasons]
        reason = ", ".join(reason_strings)
        # if the needed package is itself:
        if i == pkginfo["name"]:
            infos.append(("dependency-detected-satisfied %s (%s)", (i, reason)))
            continue
        # if the dependency is satisfied
        if i in explicitdepend or smartprovides[i] & explicitdepend:
            infos.append(("dependency-detected-satisfied %s (%s)", (i, reason)))
            continue
        # still not found, maybe it is specified as optional
        if i in optdepend or smartprovides[i] & optdepend:
            warnings.append(("dependency-detected-but-optional %s (%s)", (i, reason)))
            continue
        # a needed dependency is transitive, it is implicitly satisfied
        if i in implicitdepend or smartprovides[i] & implicitdepend:
            warnings.append(("dependency-implicitly-satisfied %s (%s)", (i, reason)))
            continue
        # maybe, it is pulled as a transitive dependency for an optdepend
        if i in implicit_optdepend or smartprovides[i] & implicit_optdepend:
            warnings.append(("dependency-implicitly-satisfied-optional %s (%s)", (i, reason)))
            continue
        # now i'm pretty sure i didn't find it.
        errors.append(("dependency-detected-not-included %s (%s)", (i, reason)))

    for i in pkginfo["depends"]:
        # multilib packages usually depend on their regular counterparts
        if pkginfo["name"].startswith("lib32-") and i == pkginfo["name"].partition("-")[2]:
            continue
        # a dependency is unneeded if:
        #   it is not in the depends as we see them
        #   it does not pull some needed dependency which provides it
        if i not in dependlist and i not in allprovides:
            warnings.append(("dependency-not-needed %s", (i,)))
    infos.append(("depends-by-namcap-sight depends=(%s)", (" ".join(dependlist),)))

    return errors, warnings, infos



# Copyright (C) 2024 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
from Namcap.tests.makepkg import MakepkgTest
import Namcap.rules.dbus1location


class Dbus1locationTest(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_dbus1location
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
source=()
build() {
  true
}
package() {
  mkdir -p "${pkgdir}/etc/dbus-1/system.d"
  touch "${pkgdir}/etc/dbus-1/system.d/foo.conf"
}
"""

    def test_dbus1location(self):
        pkgfile = "__namcap_test_dbus1location-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(
            os.path.join(self.tmpdir, pkgfile), Namcap.rules.dbus1location.dbus1locationRule
        )
        self.assertEqual(r.errors, [])
        self.assertEqual(r.warnings, [("dbus-1-location %s", "etc/dbus-1/system.d/foo.conf")])
        self.assertEqual(r.infos, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from Namcap.ruleclass import TarballRule


class package(TarballRule):
    name = "lots-of-docs"
    description = "See if a package is carrying more documentation than it should"

    def analyze(self, pkginfo, tar):
        if "name" in pkginfo and (pkginfo["name"].endswith("-doc") or pkginfo["name"].endswith("-docs")):
            # Don't do anything if the package is called "*-doc"
            return
        docdir = "usr/share/doc"
        size = 0
        docsize = 0

        for i in tar.getmembers():
            if i.name.startswith(docdir):
                docsize += i.size
            size += i.size

        if size > 0:
            ratio = docsize / float(size)
            if ratio > 0.50:
                self.warnings.append(("lots-of-docs %f", (ratio * 100,)))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os

import Namcap.rules.permissions
from Namcap.tests.makepkg import MakepkgTest


class PermissionsTest(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_permissions
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
source=()
build() {
  true
}
package() {
  install -d -m 755 ${pkgdir}/usr/bin

  touch ${pkgdir}/usr/bin/program
  chmod 777 ${pkgdir}/usr/bin/program

  touch ${pkgdir}/usr/bin/secret
  chmod 700 ${pkgdir}/usr/bin/secret

  touch ${pkgdir}/usr/bin/unsafe
  chmod u+rwxs ${pkgdir}/usr/bin/unsafe

  install -d -m 744 ${pkgdir}/usr/share/broken

  install -d -m 755 ${pkgdir}/usr/lib/libstatic.a
  ln -s libstatic.a ${pkgdir}/usr/lib/libstaticlink.a
}
"""

    def test_nonrootowner(self):
        pkgfile = "__namcap_test_permissions-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.permissions.package)
        self.assertEqual(r.errors, [])
        self.assertEqual(
            set(r.warnings),
            set(
                [
                    ("file-world-writable %s", "usr/bin/program"),
                    ("file-not-world-readable %s", "usr/bin/secret"),
                    ("file-setugid %s", "usr/bin/unsafe"),
                    ("directory-not-world-executable %s", "usr/share/broken"),
                    ("incorrect-library-permissions %s", "usr/lib/libstatic.a"),
                ]
            ),
        )
        self.assertEqual(r.infos, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import unittest

import Namcap.depends
import Namcap.package


class DependsTests(unittest.TestCase):
    def setUp(self):
        self.pkginfo = Namcap.package.PacmanPackage({"name": "package"})

    def test_missing(self):
        self.pkginfo.detected_deps = {"pkg1": []}
        e, w, i = Namcap.depends.analyze_depends(self.pkginfo)
        expected_e = [("dependency-detected-not-included %s (%s)", ("pkg1", ""))]
        self.assertEqual(e, expected_e)
        self.assertEqual(w, [])
        self.assertEqual(i, [("depends-by-namcap-sight depends=(%s)", ("pkg1",))])

    def test_unneeded(self):
        self.pkginfo["depends"] = {"pkg1": []}
        e, w, i = Namcap.depends.analyze_depends(self.pkginfo)
        expected_w = [("dependency-not-needed %s", ("pkg1",))]
        self.assertEqual(e, [])
        self.assertEqual(w, expected_w)
        self.assertEqual(i, [("depends-by-namcap-sight depends=(%s)", ("",))])

    def test_satisfied(self):
        # false negative test
        self.pkginfo["depends"] = {"readline": []}
        self.pkginfo.detected_deps = {"glibc": [], "readline": []}
        e, w, i = Namcap.depends.analyze_depends(self.pkginfo)
        expected_w = [("dependency-implicitly-satisfied %s (%s)", ("glibc", ""))]
        self.assertEqual(e, [])
        self.assertEqual(w, expected_w)
        # info is verbose and beyond scope, skip it

    def test_satisfied2(self):
        # false positive test
        self.pkginfo["depends"] = {"pyalpm": [], "python": []}
        self.pkginfo.detected_deps = {"pyalpm": [], "python": []}
        e, w, i = Namcap.depends.analyze_depends(self.pkginfo)
        self.assertEqual(e, [])
        self.assertEqual(w, [])
        # info is verbose and beyond scope, skip it



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os

import Namcap.rules.elffiles
from Namcap.tests.makepkg import MakepkgTest


class FHSELFTest(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_elffiles
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
source=()
options=(!purge !zipman)
build() {
  true
}
package() {
  install -D -m 644 /bin/ls "${pkgdir}/usr/share/something/binary"
}
"""

    def test_fhs_elffiles(self):
        pkgfile = "__namcap_test_elffiles-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.elffiles.ELFPaths)
        self.assertEqual(r.errors, [("elffile-not-in-allowed-dirs %s", ("usr/share/something/binary",))])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])


class TestExecStack(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_execstack
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
source=()
options=(!purge !zipman)
build() {
  cd "${srcdir}"
  echo "int main() { return 0; }" > main.c
  /usr/bin/gcc -o main -Wa,-execstack main.c
}
package() {
  install -D -m 644 "${srcdir}/main" "${pkgdir}/usr/bin/withexecstack"
}
"""

    def test_execstack(self):
        pkgfile = "__namcap_test_execstack-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.elffiles.ELFExecStackRule)
        self.assertEqual(r.errors, [])
        self.assertEqual(r.warnings, [("elffile-with-execstack %s", ("usr/bin/withexecstack",))])
        self.assertEqual(r.infos, [])


class TestNoPieStack(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_nopie
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
source=()
options=(!purge !zipman)
build() {
  cd "${srcdir}"
  echo "int main() { return 0; }" > main.c
  /usr/bin/gcc -o main main.c -no-pie
}
package() {
  install -D -m 644 "${srcdir}/main" "${pkgdir}/usr/bin/nopie"
}
"""

    def test_nopie(self):
        pkgfile = "__namcap_test_nopie-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.elffiles.NoPIERule)
        self.assertEqual(r.errors, [])
        self.assertEqual(r.warnings, [("elffile-nopie %s", ("usr/bin/nopie",))])
        self.assertEqual(r.infos, [])


class TestNoShadowStack(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_noshstk
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('x86_64')
url="http://www.example.com/"
license=('GPL')
depends=('glibc')
source=()
options=(!purge !zipman)
build() {
  cd "${srcdir}"
  echo "int main() { return 0; }" > main.c
  /usr/bin/gcc -o fcf-protection-branch main.c -fcf-protection=branch
  /usr/bin/gcc -o fcf-protection-full main.c -fcf-protection=full
  /usr/bin/gcc -o fcf-protection-none main.c -fcf-protection=none
  /usr/bin/gcc -o fcf-protection-none-m32 main.c -fcf-protection=none -m32 -nostdlib
  /usr/bin/gcc -o fcf-protection-return main.c -fcf-protection=return
  /usr/bin/objcopy --remove-section=.note.gnu.property fcf-protection-full no-.note.gnu.property
}
package() {
  install -D -m 644 "${srcdir}/fcf-protection-branch" "${pkgdir}/usr/bin/fcf-protection-branch"
  install -D -m 644 "${srcdir}/fcf-protection-full" "${pkgdir}/usr/bin/fcf-protection-full"
  install -D -m 644 "${srcdir}/fcf-protection-none" "${pkgdir}/usr/bin/fcf-protection-none"
  install -D -m 644 "${srcdir}/fcf-protection-none-m32" "${pkgdir}/usr/bin/fcf-protection-none-m32"
  install -D -m 644 "${srcdir}/fcf-protection-return" "${pkgdir}/usr/bin/fcf-protection-return"
  install -D -m 644 "${srcdir}/no-.note.gnu.property" "${pkgdir}/usr/bin/no-.note.gnu.property"
}
"""

    def test_noshstk(self):
        pkgfile = "__namcap_test_noshstk-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.elffiles.ELFSHSTKRule)
        self.assertEqual(r.errors, [])
        self.assertEqual(
            r.warnings,
            [
                ("elffile-noshstk %s", ("usr/bin/fcf-protection-branch",)),
                ("elffile-noshstk %s", ("usr/bin/fcf-protection-none",)),
                ("elffile-noshstk %s", ("usr/bin/no-.note.gnu.property",)),
            ],
        )
        self.assertEqual(r.infos, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import ast
import importlib
import sys
import sysconfig
from collections import defaultdict

import Namcap.package
from Namcap.ruleclass import TarballRule
from Namcap.util import is_script, script_type


def finddepends(pkgname, modules, gir_modules, gir_versions):
    """
    Find packages owning a list of libraries

    Returns:
      dependlist -- a dictionary { package => set(libraries) }
      orphans -- the list of libraries without owners
      gir_dependlist -- a dictionary { package => set(GIR libraries) }
      gir_orphans -- the list of GIR libraries without owners
    """
    python_path = sysconfig.get_path("stdlib", scheme="posix_prefix")
    site_packages_path = sysconfig.get_path("purelib", scheme="posix_prefix")

    dependlist = defaultdict(set)
    gir_dependlist = defaultdict(set)

    knownlibs: dict[str, str] = defaultdict(str)
    missinglibs = set()
    foundlibs = set()
    gir_foundlibs = set()

    for module in modules:
        # Check application-specific python modules
        if importlib.machinery.PathFinder.find_spec(
            module.split(".")[0], ["/usr/lib/" + pkgname, "/usr/share/" + pkgname]
        ):
            dependlist[pkgname].add(module)
            continue

        # Check internal python modules
        if module.split(".")[0] in sys.builtin_module_names or importlib.machinery.PathFinder.find_spec(
            module.split(".")[0], [python_path, python_path + "/lib-dynload"]
        ):
            dependlist["python"].add(module)
            continue

        # Search external python modules
        spec = importlib.machinery.PathFinder.find_spec(module.split(".")[0], [site_packages_path])
        # Search namespaced python module
        if spec and not spec.origin and spec.submodule_search_locations and len(module.split(".")) > 1:
            spec = importlib.machinery.PathFinder.find_spec(module.split(".")[1], spec.submodule_search_locations)
            if spec and not spec.origin and spec.submodule_search_locations and len(module.split(".")) > 2:
                spec = importlib.machinery.PathFinder.find_spec(module.split(".")[2], spec.submodule_search_locations)
        if spec and spec.origin:
            knownlibs[module] = spec.origin
        else:
            missinglibs.add(module)

    for pkg in Namcap.package.get_installed_packages():
        for j, fsize, fmode in pkg.files:
            if j.startswith(site_packages_path[1:]):
                for k, path in knownlibs.items():
                    if j == path[1:]:
                        dependlist[pkg.name].add(k)
                        foundlibs.add(k)

            if j.startswith("usr/lib/girepository-1.0/"):
                for module in gir_modules:
                    gir_module = module.replace("gi.repository.", "")
                    if j.startswith("usr/lib/girepository-1.0/" + gir_module + "-" + gir_versions[gir_module]):
                        gir_dependlist[pkg.name].add(module)
                        gir_foundlibs.add(module)

    orphans = list(set(knownlibs.keys()).union(missinglibs) - foundlibs)
    gir_orphans = list(set(gir_modules.keys()) - gir_foundlibs)
    return dependlist, orphans, gir_dependlist, gir_orphans


def get_imports(fileobj, filename, modules, gir_modules, gir_versions):
    try:
        root = ast.parse(fileobj.read())
    except (SyntaxError, ValueError):
        # ast.parse() uses compile(), which may raise SyntaxError or ValueError
        return

    for node in ast.walk(root):
        if isinstance(node, ast.Import):
            for module in node.names:
                modules[module.name].add(filename)
                if module.name.startswith("gi.repository."):
                    gir_modules[module.name].add(filename)
        elif isinstance(node, ast.ImportFrom):
            if node.module and node.level == 0:
                for submodule in node.names:
                    modules[node.module + "." + submodule.name].add(filename)
                    if node.module == "gi.repository":
                        gir_modules[node.module + "." + submodule.name].add(filename)
        elif (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "gi"
            and node.func.attr == "require_version"
        ):
            if hasattr(node.args[0], "value") and hasattr(node.args[1], "value"):
                gir_versions[node.args[0].value] = node.args[1].value
        elif (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "gi"
            and node.func.attr == "require_versions"
            and hasattr(node.args[0], "keys")
            and hasattr(node.args[0], "values")
        ):
            for module, version in zip(node.args[0].keys, node.args[0].values):
                if hasattr(module, "value") and hasattr(version, "value"):
                    gir_versions[module.value] = version.value


class PythonDependencyRule(TarballRule):
    name = "pydepends"
    description = "Checks python dependencies"

    def analyze(self, pkginfo, tar):
        modules: dict[str, set[str]] = defaultdict(set)
        gir_modules: dict[str, set[str]] = defaultdict(set)
        gir_versions: dict[str, str] = defaultdict(str)

        for entry in tar:
            if not entry.isfile():
                continue
            f = tar.extractfile(entry)
            if not entry.name.endswith(".py") and not is_script(f):
                continue
            if is_script(f) and script_type(f) not in ["python", "python3"]:
                continue
            get_imports(f, entry.name, modules, gir_modules, gir_versions)
            f.close()

        # If Gdk version is not defined, it should be the same as Gtk version
        if not gir_versions["Gdk"]:
            gir_versions["Gdk"] = gir_versions["Gtk"]
        if not gir_versions["GdkX11"]:
            gir_versions["GdkX11"] = gir_versions["Gtk"]

        dependlist, orphans, gir_dependlist, gir_orphans = finddepends(
            pkginfo["name"], modules, gir_modules, gir_versions
        )
        liblist = modules | gir_modules

        # Handle "no package associated" errors
        self.warnings.extend(
            [("python-module-no-package-associated %s %s", (i, str(list(liblist[i])))) for i in orphans + gir_orphans]
        )

        # Print python module deps
        for pkg, libraries in (dependlist | gir_dependlist).items():
            if isinstance(libraries, set):
                files = list(libraries)
                needing = set().union(*[liblist[lib] for lib in libraries])
                reasons = pkginfo.detected_deps.setdefault(pkg, [])
                reasons.append(("python-modules-needed %s %s", (str(files), str(list(needing)))))
                self.infos.append(("python-module-dependence %s in %s", (pkg, str(files))))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import tarfile
from contextlib import nullcontext as does_not_raise
from pathlib import Path
from tarfile import TarInfo
from typing import Any, ContextManager

from license_expression import BaseSymbol, LicenseSymbol, LicenseWithExceptionSymbol
from pytest import mark, raises

import Namcap.rules.licensepkg
from Namcap.rules import licensepkg
from Namcap.tests.makepkg import MakepkgTest


@mark.parametrize(
    "input, result, expectation",
    [
        ("LicenseRef-FOO", "LicenseRef-FOO", does_not_raise()),
        ("() broken", "", raises(Exception)),
        ("MIT and MIT", "MIT AND MIT", does_not_raise()),
        ("MIT AND Apache-2.0", "MIT AND Apache-2.0", does_not_raise()),
        (
            "Apache-2.0 or (MIT with Bootloader-exception)",
            "Apache-2.0 OR MIT WITH Bootloader-exception",
            does_not_raise(),
        ),
    ],
)
def test_get_license_canonicalized(input: str, result: str, expectation: ContextManager[Any]) -> None:
    with expectation:
        assert result == licensepkg.get_license_canonicalized(input)


@mark.parametrize(
    "input, result",
    [
        ("LicenseRef-FOO+", "LicenseRef-FOO+"),
        ("MIT", "MIT"),
        ("MIT+", "MIT"),
        ("Apache-2.0+", "Apache-2.0"),
        ("GPL-3.0-only+", "GPL-3.0-only+"),
        ("GPL-3.0-or-later+", "GPL-3.0-or-later+"),
    ],
)
def test_strip_plus_from_license(input: str, result: str) -> None:
    assert LicenseSymbol(result) == licensepkg.strip_plus_from_license(LicenseSymbol(input))


@mark.parametrize(
    "input, result, expectation",
    [
        ("Foo", {LicenseSymbol("Foo")}, does_not_raise()),
        ("() broken", set(), raises(Exception)),
        ("LicenseRef-FOO", {LicenseSymbol("LicenseRef-FOO")}, does_not_raise()),
        ("MIT AND MIT", {LicenseSymbol("MIT")}, does_not_raise()),
        ("MIT AND Apache-2.0", {LicenseSymbol("MIT"), LicenseSymbol("Apache-2.0")}, does_not_raise()),
        ("MIT+ OR Apache-2.0+", {LicenseSymbol("MIT"), LicenseSymbol("Apache-2.0")}, does_not_raise()),
        (
            "Apache-2.0 OR (MIT WITH Bootloader-exception)",
            {
                LicenseSymbol("Apache-2.0"),
                LicenseWithExceptionSymbol(
                    LicenseSymbol("MIT"), LicenseSymbol("Bootloader-exception", is_exception=True)
                ),
            },
            does_not_raise(),
        ),
        (
            "Apache-2.0+ WITH Bootloader-exception",
            {
                LicenseWithExceptionSymbol(
                    LicenseSymbol("Apache-2.0"), LicenseSymbol("Bootloader-exception", is_exception=True)
                ),
            },
            does_not_raise(),
        ),
    ],
)
def test_get_license_symbols(input: str, result: set[LicenseSymbol], expectation: ContextManager[Any]) -> None:
    with expectation:
        assert result == licensepkg.get_license_symbols(input)


@mark.parametrize(
    "input, result",
    [
        ({LicenseSymbol("GPL-3.0-or-later")}, set()),
        ({LicenseSymbol("MIT")}, {LicenseSymbol("MIT")}),
        ({LicenseSymbol("LicenseRef-FOO")}, {LicenseSymbol("LicenseRef-FOO")}),
        (
            {
                LicenseWithExceptionSymbol(
                    LicenseSymbol("MIT"), LicenseSymbol("Bootloader-exception", is_exception=True)
                )
            },
            {LicenseSymbol("MIT"), LicenseSymbol("Bootloader-exception", is_exception=True)},
        ),
    ],
)
def test_get_uncommon_license_symbols(input: set[BaseSymbol], result: set[BaseSymbol]) -> None:
    assert result == licensepkg.get_uncommon_license_symbols(
        input,
        licensepkg.get_known_spdx_license_identifiers(),
        licensepkg.get_known_spdx_license_exceptions(),
        licensepkg.get_common_spdx_license_identifiers(),
        licensepkg.get_common_spdx_license_exceptions(),
    )


@mark.parametrize(
    "input, result",
    [
        ({LicenseSymbol("GPL-3.0-or-later")}, set()),
        ({LicenseSymbol("LicenseRef-FOO")}, {LicenseSymbol("LicenseRef-FOO")}),
        ({LicenseSymbol("GPL-3.0-only+")}, {LicenseSymbol("GPL-3.0-only+")}),
        (
            {
                LicenseWithExceptionSymbol(
                    LicenseSymbol("MIT"), LicenseSymbol("Bootloader-exception", is_exception=True)
                )
            },
            set(),
        ),
    ],
)
def test_get_unknown_license_symbols(input: set[BaseSymbol], result: set[BaseSymbol]) -> None:
    assert result == licensepkg.get_unknown_license_symbols(
        input, licensepkg.get_known_spdx_license_identifiers(), licensepkg.get_known_spdx_license_exceptions()
    )


@mark.parametrize(
    "create_file, create_absolute_symlink, create_relative_symlink, create_license_dir_symlink, result",
    [
        (True, False, False, False, 1),
        (True, True, False, False, 2),
        (False, True, False, False, 1),
        (False, True, True, False, 2),
        (False, False, True, False, 1),
        (False, True, False, True, 1),
        (False, False, True, True, 1),
        (True, True, True, False, 3),
        (False, False, False, False, 0),
    ],
)
def test_package_license_files(
    create_file: bool,
    create_absolute_symlink: bool,
    create_relative_symlink: bool,
    create_license_dir_symlink: bool,
    result: int,
    tmp_path: Path,
) -> None:
    with tarfile.open(tmp_path / "test.tar.gz", "w:gz") as tar:
        tar.add(tmp_path, arcname="/usr/share/foo/")

        if create_license_dir_symlink:
            if create_absolute_symlink:
                license_dir = tmp_path / "usr/share/foo/"
                license_dir.mkdir(parents=True, exist_ok=True)
                license_file = license_dir / "LICENSE.absolute"
                license_file.touch()
                tar.add(license_file, arcname="/usr/share/foo/LICENSE.absolute")

                link_dir = tmp_path / "usr/share/licenses/"
                link_dir.mkdir(parents=True, exist_ok=True)
                link_file = link_dir / "test/"
                link_file.symlink_to(Path("/usr/share/foo/"))
                tar.add(link_file, arcname="/usr/share/licenses/test/")
            if create_relative_symlink:
                license_dir = tmp_path / "usr/share/foo/"
                license_dir.mkdir(parents=True, exist_ok=True)
                license_file = license_dir / "LICENSE.relative"
                license_file.touch()
                tar.add(license_file, arcname="/usr/share/foo/LICENSE.relative")

                link_dir = tmp_path / "usr/share/licenses/"
                link_dir.mkdir(parents=True, exist_ok=True)
                link_file = link_dir / "test/"
                link_file.symlink_to(Path("../foo/"))
                tar.add(link_file, arcname="/usr/share/licenses/test/")
        else:
            tar.add(tmp_path, arcname="/usr/share/licenses/test/")
            if create_file:
                license_file = tmp_path / "LICENSE.file"
                license_file.touch()
                tar.add(license_file, arcname="/usr/share/licenses/test/LICENSE.file")
            if create_absolute_symlink:
                license_dir = tmp_path / "usr/share/foo/"
                license_dir.mkdir(parents=True, exist_ok=True)
                license_file = license_dir / "LICENSE.absolute"
                license_file.touch()
                tar.add(license_file, arcname="/usr/share/foo/LICENSE.absolute")

                link_dir = tmp_path / "usr/share/licenses/test/"
                link_dir.mkdir(parents=True, exist_ok=True)
                link_file = link_dir / "LICENSE.absolute"
                link_file.symlink_to(Path("/usr/share/foo/LICENSE.absolute"))
                tar.add(link_file, arcname="/usr/share/licenses/test/LICENSE.absolute")
            if create_relative_symlink:
                license_dir = tmp_path / "usr/share/foo/"
                license_dir.mkdir(parents=True, exist_ok=True)
                license_file = license_dir / "LICENSE.relative"
                license_file.touch()
                tar.add(license_file, arcname="/usr/share/foo/LICENSE.relative")

                link_dir = tmp_path / "usr/share/licenses/test/"
                link_dir.mkdir(parents=True, exist_ok=True)
                link_file = link_dir / "LICENSE.relative"
                link_file.symlink_to(Path("../../foo/LICENSE.relative"))
                tar.add(link_file, arcname="/usr/share/licenses/test/LICENSE.relative")

        assert result == len(licensepkg.package_license_files(tar, "test")[0])


@mark.parametrize(
    "name, linkname, result, expectation",
    [
        ("foo", "bar", "bar", does_not_raise()),
        ("foo", "/bar", "bar", does_not_raise()),
        ("foo", "../bar", None, raises(ValueError)),
    ],
)
def test_get_symlink_target(name: str, linkname: str, result: str, expectation: ContextManager[Any]) -> None:
    tarinfo = TarInfo(name)
    tarinfo.linkname = linkname

    with expectation:
        assert result == licensepkg.get_symlink_target(tarinfo)


class LicenseFileTest(MakepkgTest):
    def test_common_license_requires_no_file(self):
        pkgbuild = """
        pkgname=__namcap_test_licensepkg
        pkgver=1.0
        pkgrel=1
        pkgdesc="A package"
        arch=('i686' 'x86_64')
        url="http://www.example.com/"
        license=('GPL-3.0-or-later')
        depends=('glibc')
        source=()
        options=(!purge !zipman)
        build() {
          true
        }
        package() {
          mkdir -p "${pkgdir}/usr/share"
          touch "${pkgdir}/usr/share/somefile"
        }
        """
        pkgfile = "__namcap_test_licensepkg-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.licensepkg.package)
        self.assertEqual(r.errors, [])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_unknown_license_identifier_without_licenseref(self):
        pkgbuild = """
            pkgname=__namcap_test_licensepkg
            pkgver=1.0
            pkgrel=1
            pkgdesc="A package"
            arch=('i686' 'x86_64')
            url="http://www.example.com/"
            license=('DWTFYWL')
            depends=('glibc')
            source=()
            options=(!purge !zipman)
            build() {
              true
            }
            package() {
              mkdir -p "${pkgdir}/usr/share"
              touch "${pkgdir}/usr/share/somefile"
            }
            """
        pkgfile = "__namcap_test_licensepkg-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.licensepkg.package)
        self.assertEqual(
            r.errors,
            [
                ("unknown-spdx-license-identifier %s", ("DWTFYWL",)),
            ],
        )
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_license_file_for_licenseref(self):
        pkgbuild = """
            pkgname=__namcap_test_licensepkg
            pkgver=1.0
            pkgrel=1
            pkgdesc="A package"
            arch=('i686' 'x86_64')
            url="http://www.example.com/"
            license=('LicenseRef-DWTFYWL')
            depends=('glibc')
            source=()
            options=(!purge !zipman)
            build() {
              true
            }
            package() {
              mkdir -p "${pkgdir}/usr/share/licenses/${pkgname}"
              touch "${pkgdir}/usr/share/licenses/${pkgname}/LICENSE"
            }
            """
        pkgfile = "__namcap_test_licensepkg-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.licensepkg.package)
        self.assertEqual(
            r.errors,
            [],
        )
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_missing_license_file_for_licenseref(self):
        pkgbuild = """
            pkgname=__namcap_test_licensepkg
            pkgver=1.0
            pkgrel=1
            pkgdesc="A package"
            arch=('i686' 'x86_64')
            url="http://www.example.com/"
            license=('LicenseRef-DWTFYWL')
            depends=('glibc')
            source=()
            options=(!purge !zipman)
            build() {
              true
            }
            package() {
              mkdir -p "${pkgdir}/usr/share"
              touch "${pkgdir}/usr/share/somefile"
            }
            """
        pkgfile = "__namcap_test_licensepkg-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.licensepkg.package)
        self.assertEqual(
            r.errors,
            [
                (
                    "license-file-missing %s %s %s",
                    ("LicenseRef-DWTFYWL", "__namcap_test_licensepkg", "0/1"),
                ),
            ],
        )
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_license_string_needs_formatting(self):
        pkgbuild = """
            pkgname=__namcap_test_licensepkg
            pkgver=1.0
            pkgrel=1
            pkgdesc="A package"
            arch=('i686' 'x86_64')
            url="http://www.example.com/"
            license=('GPL-3.0-or-later with Bootloader-exception')
            depends=('glibc')
            source=()
            options=(!purge !zipman)
            build() {
              true
            }
            package() {
              mkdir -p "${pkgdir}/usr/share"
              touch "${pkgdir}/usr/share/somefile"
            }
            """
        pkgfile = "__namcap_test_licensepkg-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.licensepkg.package)
        self.assertEqual(
            r.errors,
            [
                (
                    "license-statement-formatting %s %s",
                    ("GPL-3.0-or-later with Bootloader-exception", "GPL-3.0-or-later WITH Bootloader-exception"),
                ),
                ("license-file-missing %s %s %s", ("Bootloader-exception", "__namcap_test_licensepkg", "0/1")),
            ],
        )
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_invalid_license_string(self):
        pkgbuild = """
            pkgname=__namcap_test_licensepkg
            pkgver=1.0
            pkgrel=1
            pkgdesc="A package"
            arch=('i686' 'x86_64')
            url="http://www.example.com/"
            license=('() broken')
            depends=('glibc')
            source=()
            options=(!purge !zipman)
            build() {
              true
            }
            package() {
              mkdir -p "${pkgdir}/usr/share"
              touch "${pkgdir}/usr/share/somefile"
            }
            """
        pkgfile = "__namcap_test_licensepkg-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.licensepkg.package)
        self.assertEqual(
            r.errors,
            [
                (
                    "invalid-license-string %s",
                    ("() broken",),
                ),
            ],
        )
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])

    def test_symlink_to_missing_file_in_other_pkg(self):
        pkgbuild = """
        pkgname=__namcap_test_licensepkg
        pkgver=1.0
        pkgrel=1
        pkgdesc="A package"
        arch=('i686' 'x86_64')
        url="http://www.example.com/"
        license=('MIT')
        depends=('glibc')
        source=()
        options=(!purge !zipman)
        build() {
          true
        }
        package() {
          mkdir -p "${pkgdir}/usr/share/licenses/__namcap_test_licensepkg/"
          ln -sf /usr/share/licenses/other_pkg/LICENSE "${pkgdir}/usr/share/licenses/__namcap_test_licensepkg/"
        }
        """
        pkgfile = "__namcap_test_licensepkg-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.licensepkg.package)
        self.assertEqual(
            r.errors,
            [
                (
                    "license-file-missing-in-other-pkg %s",
                    ("usr/share/licenses/other_pkg/LICENSE",),
                ),
            ],
        )
        self.assertEqual(r.warnings, [("license-file-in-external-pkg %s", ("usr/share/licenses/other_pkg/LICENSE",))])
        self.assertEqual(r.infos, [])

    def test_license_dir_is_symlink_and_target_pkg_not_in_depends(self):
        pkgbuild = """
        pkgname=__namcap_test_licensepkg
        pkgver=1.0
        pkgrel=1
        pkgdesc="A package"
        arch=('i686' 'x86_64')
        url="http://www.example.com/"
        license=('MIT')
        depends=('glibc')
        source=()
        options=(!purge !zipman)
        build() {
          true
        }
        package() {
          mkdir -p "${pkgdir}/usr/share/licenses/"
          ln -sf /usr/share/licenses/other_pkg/ "${pkgdir}/usr/share/licenses/__namcap_test_licensepkg"
        }
        """
        pkgfile = "__namcap_test_licensepkg-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.licensepkg.package)
        self.assertEqual(
            r.errors,
            [
                ("license-dir-target-pkg-not-in-depends %s %s %s", ("other_pkg", "MIT", "0/1")),
            ],
        )
        self.assertEqual(r.warnings, [("license-dir-in-external-pkg %s", ("usr/share/licenses/other_pkg",))])
        self.assertEqual(r.infos, [])

    def test_license_dir_is_symlink_and_dependencies_do_not_provide_files(self):
        pkgbuild = """
        pkgname=__namcap_test_licensepkg
        pkgver=1.0
        pkgrel=1
        pkgdesc="A package"
        arch=('i686' 'x86_64')
        url="http://www.example.com/"
        license=('MIT')
        depends=('licenses')
        source=()
        options=(!purge !zipman)
        build() {
          true
        }
        package() {
          mkdir -p "${pkgdir}/usr/share/licenses/"
          ln -sf /usr/share/foo/ "${pkgdir}/usr/share/licenses/__namcap_test_licensepkg"
        }
        """
        pkgfile = "__namcap_test_licensepkg-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(os.path.join(self.tmpdir, pkgfile), Namcap.rules.licensepkg.package)
        self.assertEqual(
            r.errors,
            [
                (
                    "license-dir-is-symlink-and-license-files-missing %s %s %s",
                    ("MIT", "usr/share/foo", "0/1"),
                ),
            ],
        )
        self.assertEqual(r.warnings, [("license-dir-in-external-pkg %s", ("usr/share/foo",))])
        self.assertEqual(r.infos, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import Namcap.rules.splitpkgbuild as module
from Namcap.tests.pkgbuild_test import PkgbuildTest


class NamcapSplitPkgbuildTest(PkgbuildTest):
    pkgbuild1 = """
# Maintainer: Arch Linux <archlinux@example.com>
# Contributor: Arch Linux <archlinux@example.com>

pkgbase=mypackage
pkgname=('prog1' 'prog2')
pkgver=1.0
pkgrel=1
pkgdesc="A package"
url="http://mypackage.sf.net/"
arch=('i686' 'x86_64')
depends=('glibc')
license=('GPL-3.0-or-later')
options=('!libtool')
source=(ftp://heanet.dl.sourceforge.net/pub/m/mypackage/mypackage-0.1.tar.gz)
md5sums=('abcdefabcdef12345678901234567890')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make
}

package_prog1() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make DESTDIR="${pkgdir}" install1
}
"""

    test_valid = PkgbuildTest.valid_tests

    def preSetUp(self):
        self.rule = module.PackageFunctionsRule

    def test_example1(self):
        "Example 1 : missing package_* function"
        r = self.run_on_pkg(self.pkgbuild1)
        self.assertEqual(r.errors, [("missing-pkgfunction %s", ("prog2",))])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])


class NamcapSplitMakedepsTest(PkgbuildTest):
    pkgbuild1 = """
# Maintainer: Arch Linux <archlinux@example.com>
# Contributor: Arch Linux <archlinux@example.com>

pkgbase=mypackage
pkgname=('prog1' 'prog2')
pkgver=1.0
pkgrel=1
pkgdesc="A package"
url="http://mypackage.sf.net/"
arch=('i686' 'x86_64')
makedepends=('gtk2')
license=('GPL-3.0-or-later')
options=('!libtool')
source=(ftp://heanet.dl.sourceforge.net/pub/m/mypackage/mypackage-0.1.tar.gz)
md5sums=('abcdefabcdef12345678901234567890')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make
}

package_prog1() {
  depends=("gtk2")
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make DESTDIR="${pkgdir}" install1
}

package_prog2() {
  depends=("lib1" "lib2")
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make DESTDIR="${pkgdir}" install1
}
"""

    test_valid = PkgbuildTest.valid_tests

    def preSetUp(self):
        self.rule = module.SplitPkgMakedepsRule

    def test_example1(self):
        "Example 1: missing makedepend"
        r = self.run_on_pkg(self.pkgbuild1)
        self.assertEqual(r.errors, [("missing-makedeps %s", (str(["lib1", "lib2"]),))])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os

import Namcap.rules.pydepends
from Namcap.tests.makepkg import MakepkgTest


class PyDependsTest(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_pydepends
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('any')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('python')
source=()
build() {
  cd "${srcdir}"
  echo "import importlib, pyalpm" > main.py
  echo "(^y^)" > invalid.py
}
package() {
  install -D -m 755 "$srcdir/main.py" "$pkgdir/usr/bin/main.py"
  install -D -m 755 "$srcdir/invalid.py" "$pkgdir/usr/bin/invalid.py"
}
"""

    def test_pydepends(self):
        "Package with missing pacman dependency"
        pkgfile = "__namcap_test_pydepends-1.0-1-any.pkg.tar"
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(
            os.path.join(self.tmpdir, pkgfile), Namcap.rules.pydepends.PythonDependencyRule
        )
        self.assertEqual(
            pkg.detected_deps["pyalpm"], [("python-modules-needed %s %s", (str(["pyalpm"]), str(["usr/bin/main.py"])))]
        )
        e, w, i = Namcap.depends.analyze_depends(pkg)
        self.assertEqual(
            e,
            [
                (
                    "dependency-detected-not-included %s (%s)",
                    ("pyalpm", "python modules ['pyalpm'] needed in files ['usr/bin/main.py']"),
                )
            ],
        )
        self.assertEqual(w, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from types import ModuleType

import Namcap.ruleclass

# Tarball rules
# PKGBUILD and metadata rules
from . import (  # noqa: F401
    anyelf,
    arrays,
    badbackups,
    carch,
    dbus1location,
    elffiles,
    emptydir,
    externalhooks,
    extravars,
    fhs,
    filenames,
    fileownership,
    gnomemime,
    hardlinks,
    hookdepends,
    infodirectory,
    invalidstartdir,
    javafiles,
    libtool,
    licensepkg,
    lotsofdocs,
    makedepends,
    makepkgfunctions,
    missingbackups,
    missingvars,
    pathdepends,
    pcdepends,
    perllocal,
    permissions,
    pkginfo,
    pkgnameindesc,
    py_mtime,
    pydepends,
    qmldepends,
    rpath,
    runpath,
    scrollkeeper,
    sfurl,
    shebangdepends,
    sodepends,
    sphinxbuildcachefiles,
    splitpkgbuild,
    symlink,
    systemdlocation,
    unusedsodepends,
)

all_rules = {}
for name, value in dict(locals()).items():
    if not isinstance(value, ModuleType):
        continue
    if name == "Namcap.ruleclass":
        continue
    for n, v in value.__dict__.items():
        if isinstance(v, type) and issubclass(v, Namcap.ruleclass.AbstractRule) and hasattr(v, "name"):
            all_rules[v.name] = v



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os

tags: dict[str, str] = {}

DEFAULT_TAGS = "/usr/share/namcap/namcap-tags"


def load_tags(filename=None, machine=False):
    "Loads tags from the given filename"
    global tags
    tags = {}
    if filename is None:
        filename = DEFAULT_TAGS

    f = open(filename)
    for line in f:
        line = line.strip()
        if line.startswith("#") or line == "":
            continue
        machinetag, humantag = line.split("::")
        machinetag = machinetag.strip()
        humantag = humantag.strip()

        if machine:
            tags[machinetag] = machinetag
        else:
            tags[machinetag] = humantag


def format_message(msg):
    """
    Formats a tuple (tag, data)
    """
    tag, data = msg
    return tags[tag] % data


# Try to load tags by default
if os.path.exists(DEFAULT_TAGS):
    load_tags(DEFAULT_TAGS)
elif os.path.exists("namcap-tags"):
    load_tags("namcap-tags")



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from Namcap.ruleclass import TarballRule


class package(TarballRule):
    name = "gnomemime"
    description = "Checks for generated GNOME mime files"

    def analyze(self, pkginfo, tar):
        mime_files = [
            "usr/share/applications/mimeinfo.cache",
            "usr/share/mime/XMLnamespaces",
            "usr/share/mime/aliases",
            "usr/share/mime/globs",
            "usr/share/mime/magic",
            "usr/share/mime/subclasses",
        ]

        for i in tar.getnames():
            if i in mime_files:
                self.errors.append(("gnome-mime-file %s", i))



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import re
import subprocess
import tempfile

from Namcap.ruleclass import TarballRule
from Namcap.util import is_elf

libre = re.compile(r"^\t(/.*)")
lddfail = re.compile(r"^\tnot a dynamic executable")


def get_unused_sodepends(filename):
    p = subprocess.Popen(
        ["ldd", "-r", "-u", filename], env={"LANG": "C"}, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    var = p.communicate()
    if p.returncode == 0:
        return
    for j in var[0].decode("ascii").splitlines():
        # Don't raise an error, as the executable might be a valid ELF file,
        # just not a dynamically linked one
        n = lddfail.search(j)
        if n is not None:
            return

        n = libre.search(j)
        if n is not None:
            yield n.group(1)


class package(TarballRule):
    name = "unusedsodepends"
    description = "Checks for unused dependencies caused by linked shared libraries"

    def analyze(self, pkginfo, tar):
        for entry in tar:
            if not entry.isfile():
                continue

            # is it an ELF file ?
            f = tar.extractfile(entry)
            if not is_elf(f):
                f.close()
                continue
            elf = f.read()
            f.close()

            # write it to a temporary file
            f = tempfile.NamedTemporaryFile(delete=False)
            f.write(elf)
            f.close()

            os.chmod(f.name, 0o755)

            for lib in get_unused_sodepends(f.name):
                self.warnings.append(("unused-sodepend %s %s", (lib, entry.name)))

            os.unlink(f.name)



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

from elftools.elf.dynamic import DynamicSection
from elftools.elf.elffile import ELFFile
from elftools.elf.enums import ENUM_GNU_PROPERTY_X86_FEATURE_1_FLAGS
from elftools.elf.sections import NoteSection, SymbolTableSection

from Namcap.ruleclass import TarballRule
from Namcap.util import is_elf

# Valid directories for ELF files
valid_dirs = ["bin/", "sbin/", "usr/bin/", "usr/sbin/", "lib/", "usr/lib/", "usr/lib32/"]
# Questionable directories for ELF files
# (Suppresses some output spam.)
questionable_dirs = ["opt/"]


def elf_files_from_tar(tar):
    for entry in tar:
        if not entry.isfile():
            continue
        fileobj = tar.extractfile(entry)
        if not fileobj or not is_elf(fileobj):
            continue
        yield ELFFile(fileobj), entry.name


class ELFPaths(TarballRule):
    name = "elfpaths"
    description = "Check about ELF files outside some standard paths."

    def analyze(self, pkginfo, tar):
        invalid_elffiles = []
        questionable_elffiles = []

        for entry in tar:
            # is it a regular file ?
            if not entry.isfile():
                continue
            # is it outside standard binary dirs ?
            in_std_dirs = any(entry.name.startswith(d) for d in valid_dirs)
            in_que_dirs = any(entry.name.startswith(d) for d in questionable_dirs)

            if in_std_dirs:
                continue
            # is it an ELF file ?
            f = tar.extractfile(entry)
            if is_elf(f):
                if in_que_dirs:
                    questionable_elffiles.append(entry.name)
                else:
                    invalid_elffiles.append(entry.name)

        que_elfdirs = [d for d in questionable_dirs if any(f.startswith(d) for f in questionable_elffiles)]
        self.errors = [("elffile-not-in-allowed-dirs %s", (i,)) for i in invalid_elffiles]
        self.errors.extend(("elffile-in-questionable-dirs %s", (i,)) for i in que_elfdirs)
        self.infos = [("elffile-not-in-allowed-dirs %s", (i,)) for i in questionable_elffiles]


class ELFTextRelocationRule(TarballRule):
    """
    Check for text relocations in ELF files.
    """

    name = "elftextrel"
    description = "Check for text relocations in ELF files."

    def analyze(self, pkginfo, tar):
        files_with_textrel = []

        for elffile, entry_name in elf_files_from_tar(tar):
            for section in elffile.iter_sections():
                if not isinstance(section, DynamicSection):
                    continue
                for tag in section.iter_tags("DT_TEXTREL"):
                    if tag.entry.d_tag == "DT_TEXTREL":
                        files_with_textrel.append(entry_name)

        if files_with_textrel:
            self.warnings = [("elffile-with-textrel %s", (i,)) for i in files_with_textrel]


class ELFExecStackRule(TarballRule):
    """
    Check for executable stacks in ELF files.

    Introduced by FS#26458. Uses pyelftools to read the GNU_STACK
    program header and ensure it does not have the executable bit
    set.
    """

    name = "elfexecstack"
    description = "Check for executable stacks in ELF files."

    def analyze(self, pkginfo, tar):
        exec_stacks = []

        for elffile, entry_name in elf_files_from_tar(tar):
            for segment in elffile.iter_segments():
                if segment["p_type"] != "PT_GNU_STACK":
                    continue

                mode = segment["p_flags"]
                if mode & 1:
                    exec_stacks.append(entry_name)

        if exec_stacks:
            self.warnings = [("elffile-with-execstack %s", (i,)) for i in exec_stacks]


class ELFGnuRelroRule(TarballRule):
    """
    Check for read-only relocation in ELF files.

    Introduced by FS#26435. Uses pyelftools to check for GNU_RELRO.
    """

    name = "elfgnurelro"
    description = "Check for FULL RELRO in ELF files."

    def has_bind_now(self, elffile):
        DF_BIND_NOW = 0x08

        for section in elffile.iter_sections():
            if not isinstance(section, DynamicSection):
                continue
            for tag in section.iter_tags():
                if tag.entry.d_tag == "DT_BIND_NOW":
                    return True
                if tag.entry.d_tag == "DT_FLAGS" and tag.entry.d_val & DF_BIND_NOW:
                    return True
        return False

    def analyze(self, pkginfo, tar):
        missing_relro = []

        for elffile, entry_name in elf_files_from_tar(tar):
            if ".debug" in entry_name:
                continue

            if any(seg["p_type"] == "PT_GNU_RELRO" for seg in elffile.iter_segments()):
                if self.has_bind_now(elffile):
                    continue

            missing_relro.append(entry_name)

        if missing_relro:
            self.warnings = [("elffile-without-relro %s", (i,)) for i in missing_relro]


class ELFUnstrippedRule(TarballRule):
    """
    Checks for unstripped ELF files. Uses pyelftools to check if
    .symtab exists.

    Introduced by FS#27485.
    """

    name = "elfunstripped"
    description = "Check for unstripped ELF files."

    def analyze(self, pkginfo, tar):
        unstripped_binaries = []

        for elffile, entry_name in elf_files_from_tar(tar):
            if ".debug" in entry_name:
                continue
            for section in elffile.iter_sections():
                if not isinstance(section, SymbolTableSection):
                    continue

                if section["sh_entsize"] == 0:
                    continue

                if section.name == ".symtab":
                    unstripped_binaries.append(entry_name)
        if unstripped_binaries:
            self.warnings = [("elffile-unstripped %s", (i,)) for i in unstripped_binaries]


class NoPIERule(TarballRule):
    """
    Checks for no PIE ELF files.
    """

    name = "elfnopie"
    description = "Check for no PIE ELF files."

    def has_dt_debug(self, elffile):
        for section in elffile.iter_sections():
            if not isinstance(section, DynamicSection):
                continue
            for _ in section.iter_tags("DT_DEBUG"):
                return True
        return False

    def analyze(self, pkginfo, tar):
        nopie_binaries = []

        for elffile, entry_name in elf_files_from_tar(tar):
            if any(x in entry_name for x in [".so", ".debug"]):
                continue
            if elffile.header["e_type"] != "ET_DYN" or not self.has_dt_debug(elffile):
                nopie_binaries.append(entry_name)

        if nopie_binaries:
            self.warnings = [("elffile-nopie %s", (i,)) for i in nopie_binaries]


def _note_props(elffile, note_type, prop_type):
    for section in elffile.iter_sections():
        if not isinstance(section, NoteSection):
            continue
        for note in section.iter_notes():
            if note["n_type"] != note_type:
                continue
            for prop in note["n_desc"]:
                if prop["pr_type"] == prop_type:
                    yield prop


class ELFSHSTKRule(TarballRule):
    """
    Check shadow stack support in ELF files.
    """

    #  This lint functions properly, but given that many/most language
    #  ecosystems don't have support yet this turns up more too many false
    #  positives to be useful. Enable at a later date when tooling supports it
    #  and our warning message points in the right direction to make it work.
    enable = False

    name = "elfnoshstk"
    description = "Check for shadow stack support in ELF files."

    def analyze(self, pkginfo, tar):
        noshstk_binaries = []

        for elffile, entry_name in elf_files_from_tar(tar):
            if ".debug" in entry_name or elffile["e_machine"] != "EM_X86_64":
                continue
            for prop in _note_props(
                elffile,
                note_type="NT_GNU_PROPERTY_TYPE_0",
                prop_type="GNU_PROPERTY_X86_FEATURE_1_AND",
            ):
                if prop["pr_data"] & ENUM_GNU_PROPERTY_X86_FEATURE_1_FLAGS["GNU_PROPERTY_X86_FEATURE_1_SHSTK"]:
                    break

            else:
                noshstk_binaries.append(entry_name)
        if noshstk_binaries:
            self.warnings = [("elffile-noshstk %s", (i,)) for i in noshstk_binaries]



# Copyright (C) 2024 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
from Namcap.tests.makepkg import MakepkgTest
import Namcap.rules.sphinxbuildcachefiles


class SphinxBuildCacheFilesTest(MakepkgTest):
    pkgbuild = """
pkgname=__namcap_test_sphinxbuildcachefiles
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL-3.0-or-later')
depends=('glibc')
source=()
build() {
  true
}
package() {
  touch ${pkgdir}/environment.pickle
}
"""

    def test_cache_files(self):
        pkgfile = "__namcap_test_sphinxbuildcachefiles-1.0-1-%(arch)s.pkg.tar" % {"arch": self.arch}
        with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
            f.write(self.pkgbuild)
        self.run_makepkg()
        pkg, r = self.run_rule_on_tarball(
            os.path.join(self.tmpdir, pkgfile), Namcap.rules.sphinxbuildcachefiles.sphinxbuildcachefilesRule
        )
        self.assertEqual(r.errors, [("sphinx-build-cache-files", ())])
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.infos, [])



# Copyright (C) 2003-2023 Namcap contributors, see AUTHORS for details.
# SPDX-License-Identifier: GPL-2.0-or-later

"""Verifies that no specific host type is used"""

import re

from Namcap.ruleclass import PkgbuildRule


class package(PkgbuildRule):
    name = "carch"
    description = "Verifies that no specific host type is used"

    def analyze(self, pkginfo, tar):
        arches = ["arm", "i586", "i686", "ppc", "x86_64"]
        archmatch = re.compile(r"\b(%s)\b" % "|".join(arches))
        # Match an arch=(i686) line
        archline = re.compile(r"\w*arch=")
        for line in pkginfo.pkgbuild:
            # strip any comments
            line = line.split("#")[0]
            match = archmatch.search(line)
            if not match:
                continue
            if archline.match(line):
                continue
            if "$CARCH" in line:
                continue
            if "${CARCH}" in line:
                continue
            self.warnings.append(("specific-host-type-used %s", (match.group(1),)))
